# -*- coding:utf-8 -*-

"""PyHdust *input* module: Hdust input tools.

Definitions on Classes:
- fname: is the original file for reference (kept as stated in the input). 
The following variables are defined based on the fname.
- proj: directory to the project (mandatory)
- modn: 'modn' structure
- suf: only the suffix.
- one use Class.set_fname(fname) to update everything.
- convention is to use "_" as separador. So, use "-" for a given **name**, as 
for the projects.

:co-author: Rodrigo Vieira
:license: GNU GPL v3.0 https://github.com/danmoser/pyhdust/blob/master/LICENSE
"""
from __future__ import print_function
import os as _os
import re as _re
import numpy as _np
from glob import glob as _glob
from itertools import product as _product
import pyhdust.phc as _phc
import pyhdust.rotstars as _rot
import pyhdust as _hdt
from collections import OrderedDict as _OrderedDict
from collections import Mapping as _mapdict
from six import string_types as _string_types
import warnings as _warn

__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


class Input(object):
    """ Input class; no fname (ie., specific suf/model defined) 

    - Criado no novo `make_inpjob`. 
    - Terminando verificação de existência prévia de *.inp e *.sed2
    - Incluído verificação do *_SEI.sed2
    """
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_disco.txt'))
    mod = f0.readlines()
    f0.close()

    attl_to_chk = ['case1', 'case2', 'composition', 'controls', 'gridcells', 
        'observers', 'perturbations', 'source']

    def __init__(self, proj=None, modn='mod01'):
        self.proj = proj
        if proj is None:
            self.proj = _os.path.split(_os.getcwd())[1]
        self.set_modn(modn)
        return

    def set_modn(self, modn):
        self.modn = modn
        if isinstance(modn, _string_types):
            if modn.isdigit():
                modn = int(modn)
            elif not modn.startswith('mod'):
                raise ValueError('# ERROR! modn must startswith "mod"')
        if isinstance(modn, ( int, long )):
            if modn > 0 and modn < 1000:
                self.modn = 'mod{0:02d}'.format(int(modn))
            else:
                raise ValueError('# ERROR! modn must be <1000')

    def set_input(self, docasesl=None, case1='step1', case2='step1_refine', 
        case3l=None, imagesl=None, composition='pureH', controls='controls', 
        gridcells='grid', observers='observers', perturbations='', source='', 
        chkout=True, c1max=20, c2max=24, c3prefl=None, clustl=None, ncore=48, 
        walltime='72:00:00', email='$USER@localhost', touch=False, **kwargs):

        if isinstance(case3l, _string_types):
            self.case3l = [case3l]
        elif case3l is None:
            self.case3l = []
        else:
            self.case3l = case3l
        if isinstance(imagesl, _string_types):
            self.imagesl = [imagesl]
        elif imagesl is None:
            self.imagesl = []
        else:
            self.imagesl = imagesl

        # docases == list of numbers    
        if docasesl is None:
            self.docasesl = [1, 2, 3]
        else:
            self.docasesl = docasesl

        self.case1 = case1
        self.case2 = case2
        self.composition = composition
        self.controls = controls
        self.gridcells = gridcells
        self.observers = observers 
        self.perturbations = perturbations
        self.source = source

        self.chkout = chkout
        self.c1max = str(c1max)
        self.c2max = str(c2max)
        if isinstance(c3prefl, _string_types):
            self.c3prefl = [c3prefl]
        elif c3prefl is None:
            self.c3prefl = []
        else:
            self.c3prefl = c3prefl
        if isinstance(clustl, _string_types):
            self.clustl = [clustl]
        elif clustl is None:
            self.clustl = []
        else:
            self.clustl = clustl
        self.ncore = ncore
        self.walltime = walltime
        self.email = email
        self.touch = touch

        for it in kwargs.items():
            setattr(self, *it)
        return

    def write_inp(self):
        bdir = self.proj
        if bdir == _os.path.split(_os.getcwd())[1]:
            bdir = ""

        head = "PROJECT = {0}\nMODEL = {1}".format(_os.path.split(
            self.proj)[1], self.modn[3:])

        c1 = "\n\n"
        if (_os.path.exists(_os.path.join(bdir, self.modn, self.modn + 
            self.suf + self.c1max + '.temp')) and self.chkout) or (1 not in 
            self.docasesl):
                c1 += '! '
        c1 += ("SUFFIX='{suf}'  SIMULATION='{case1}' SOURCE='{source}' "
            "COMPOSITION='{composition}' GRIDCELLS='{gridcells}' "
            "CONTROLS='{controls}' ".format(suf=self.suf, case1=self.case1,
                source=self.source, composition=self.composition, 
                gridcells=self.gridcells, controls=self.controls))
        if len(self.perturbations) > 0:
            c1 += "PERTURBATIONS='{0}' ".format(self.perturbations)

        c2 = '\n\n'
        if (_os.path.exists(_os.path.join(bdir, self.modn, self.modn + 
            self.suf + self.c2max + '.temp')) and self.chkout) or (2 not in 
            self.docasesl):
                c2 += '! '
        c2 += ("SUFFIX='{suf}'  SIMULATION='{case2}' SOURCE='{source}' "
            "COMPOSITION='{composition}' GRIDCELLS='{gridcells}' "
            "CONTROLS='{controls}' ".format(suf=self.suf, case2=self.case2,
                source=self.source, composition=self.composition, 
                gridcells=self.gridcells, controls=self.controls))
        if len(self.perturbations) > 0:
            c2 += "PERTURBATIONS='{0}' ".format(self.perturbations)

        c3 = '\n\n'
        gen = (sim for sim in self.case3l if 3 in self.docasesl)
        for sim in gen:
            if self.chkout:
                pref = self.c3prefl[self.case3l.index(sim)]
                if not pref.endswith('_'):
                    pref = pref+'_'
                s2name = _os.path.join(bdir, self.modn, pref + self.modn + 
                    self.suf)
                if (_os.path.exists(s2name + '.sed2')) or (
                    _os.path.exists(s2name + '_SEI.sed2')):
                    c3 += '! '
            c3 += ("SUFFIX='{suf}'  SIMULATION='{sim}' SOURCE='{source}' "
                "COMPOSITION='{composition}' GRIDCELLS='{gridcells}' "
                "CONTROLS='{controls}' OBSERVERS='{obs}' ".format(suf=self.suf, 
                    sim=sim, source=self.source, 
                    composition=self.composition, gridcells=self.gridcells, 
                    controls=self.controls, obs=self.observers))
            if len(self.perturbations) > 0:
                c3 += "PERTURBATIONS='{0}' ".format(self.perturbations)
            img = self.imagesl[self.case3l.index(sim)]
            if len(img) > 0:
                c3 += "IMAGES='{0}' ".format(img)
            c3 += '\n'

        outname = _os.path.join(bdir, self.modn, self.modn + self.suf + '.inp')
        f0 = open(outname, 'w')
        f0.writelines(head+c1+c2+c3)
        f0.close()
        print('# {0} written!'.format(outname))
        return 

    def _hassiminp(self, inpname):
        inp = open(inpname).read().split('\n')
        sims = [1 for line in inp if (line.upper().find('SUFFIX')>= 0 and 
            not line.startswith('!'))]
        return bool(sims)

    def write_job(self):
        if self.touch is True:
            raise NotImplementedError('# touch option not available')

        bdir = self.proj
        if bdir == _os.path.split(_os.getcwd())[1]:
            bdir = ""
        outname = _os.path.join(bdir, self.modn, self.modn + self.suf + '.inp')
        intname = _os.path.join(self.proj, self.modn, 
            self.modn + self.suf + '.inp')

        if not self._hassiminp(outname):
            print('# Message: No simulations inside {0}'.format(outname))
            return

        for cl in self.clustl:
            if cl.lower().startswith('job'):
                cl = 'job'
                reffile = 'REF.job'
                reps = ( (3, 'hdust', 'hd_'+'_'.join((self.proj, self.modn))), 
                    (4, '128', self.ncore), (4, '36:00:00', self.walltime), 
                    (8, 'alexcarciofi@gmail.com', self.email), 
                    (11, 'hdust_bestar2.02.inp', intname),
                    (23, 'cd ${PBS_O_WORKDIR}', ('cd ${{PBS_O_WORKDIR}}\n'
                        'printf "{0}\\n\\n" >> output_${{PBS_JOBID}}\n'.format(
                            intname))),
                )
            elif cl.lower().startswith('oar'):
                cl = 'oar'
                reffile = 'REF.oar'
                reps = ( (1, 'hdust_dmf', 'hd_'+'_'.join((self.proj, 
                    self.modn))), 
                    (2, '12', int(round(self.ncore/12.))), 
                    (2, '24:00:00', self.walltime), 
                    (10, 'hdust_bestar2.02.inp', '{0}\n\nprintf "{0}\\n\\n"'
                        ' >> $OAR_JOB_ID'.format(intname)),
                )
            else:
                _warn.warn('# Warning! Option {0} for clusters ignored'.format(
                    cl))
                continue

            if not _os.path.exists(_os.path.join(bdir, cl)):
                _os.makedirs(_os.path.join(bdir, cl))
            clname = (_os.path.join(bdir, cl, self.modn + self.suf)+"."+cl)

            f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', reffile))
            modi = f0.readlines()
            f0.close()
            for r in reps:
                _phc.repl_fline_val(modi, *r)

            f0 = open(clname, 'w')
            print('# {0} written!'.format(clname))
            f0.writelines(modi)
            f0.close()
        return


class Disk(Input):
    """ Disk class """
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_disco.txt'))
    mod = f0.readlines()
    f0.close()

    def __init__(self, proj=None, modn='mod01'):
        super(Disk, self).__init__(proj=proj, modn=modn)
        return

    def set_disk(self, renv=18.6, mh=1.5, ht=60., nr=2., dval=1e12, hseq=False, 
        alpha=.5, mu=.5, R0r=100, vt=0., denstype=None, **kwargs):
        """ denstype = ['n0', 'sig0', 'mdot'] 

        ``sig0`` means g/cm2 instead of ``n0``"""
        self.renv = renv
        self.mh = mh
        self.ht = ht
        self.nr = nr
        self.dval = dval
        self.hseq = hseq
        self.alpha = alpha
        self.mu = mu
        self.R0r = R0r
        self.vt = vt
        if isinstance(denstype, _string_types):
            if denstype.lower() in ['n0', 'sig0', 'mdot']:
                self.denstype = denstype
            denstype = None
        if denstype is None:
            _warn.warn('# Warning! Guessing the denstype...')
            self.denstype = 'n0'
            if dval < 1e-2:
                self.denstype = 'mdot'
            elif dval < 1e2:
                self.denstype = 'sig0'

        for it in kwargs.items():
            setattr(self, *it)
        return

    def set_suf(self, vfmt=''):
        """    vfmt = dict(zip(['renv', 'dval', 'ht', 'M', "n"], 
        ['{:04.1f}', '{:.1e}', '{:02.0f}', '{:04.1f}', '{:.1f}'])) """
        if isinstance(vfmt, _mapdict):
            self.vfmt = vfmt
        if hasattr(self, 'vfmt'):
            if isinstance(self.vfmt, _mapdict):    
                self.suf = '_' + '_'.join(it[0]+it[1].format(getattr(self, 
                    it[0])) for it in self.vfmt.items())
            else:
                delattr(self, vfmt)
                self.suf = vfmt
        else:
            self.suf = vfmt

    def __str__(self):
        return self.suf

    def makedisk(self):
        if not self.modn.startswith('mod'):
            raise ValueError('# ERROR! Invalid Disk.modn')
        modi = Disk.mod[:]

        if self.hseq is True and self.denstype is not 'mdot':
            print('# Warning! Your choice of HSEQ don\'t appears to be '
                'consistent')

        _ = _phc.repl_fline_val(modi, 13, '18.6', self.renv)
        if self.denstype != 'mdot':
            _ = _phc.repl_fline_val(modi, 20, '2.0', self.nr)
        else:
            _ = _phc.repl_fline_val(modi, 23, '1.', self.alpha)
            _ = _phc.repl_fline_val(modi, 24, '0.', self.R0r)
        if self.hseq:
            _ = _phc.repl_fline_val(modi, 25, '= 0', '= 1')
            _ = _phc.repl_fline_val(modi, 31, '0', '1')
            _ = _phc.repl_fline_val(modi, 36, '1.5', self.mh)
        else:
            _ = _phc.repl_fline_val(modi, 33, '1.5', self.mh)
        if float(self.ht) > 1000:
            _ = _phc.repl_fline_val(modi, 40, '18000.', self.ht)
        else:
            _ = _phc.repl_fline_val(modi, 38, '1', '2')
            if self.ht < 1:
                self.ht *= 100
            _ = _phc.repl_fline_val(modi, 43, '72.', self.ht)

        if self.denstype == 'sig0':
            raise ValueError('# ERROR! `sig0` is not a valid option yet')
        elif self.denstype == 'n0':
            _ = _phc.repl_fline_val(modi, 52, '2.35E13', self.dval)
        else:
            _ = _phc.repl_fline_val(modi, 49, '2', '3')
            _ = _phc.repl_fline_val(modi, 55, '1.E-9', self.dval)

        _ = _phc.repl_fline_val(modi, 63, '0.', self.vt)

        bdir = self.proj
        if bdir == _os.path.split(_os.getcwd())[1]:
            bdir = ""
        if not _os.path.exists(_os.path.join(bdir, self.modn)):
            _os.makedirs(_os.path.join(bdir, self.modn))
        self.fname = (_os.path.join(bdir, self.modn, self.modn + self.suf) + 
            ".txt")
        f0 = open(self.fname, 'w')
        print('# {0} written!'.format(self.fname))
        f0.writelines(modi)
        f0.close()

        return


class HdustMod(object):
    """ HdustMod doc 
    """
    def __init__(self, fname=None):
        self.fname = fname
        if fname is not None:
            self.set_fname(fname)
        return

    def set_fname(self, fname):
        self.fname = fname
        self.modn, self.suf = _os.path.split(fname)
        self.suf = _os.path.splitext(self.suf)[0]
        if (not self.modn.startswith('mod')) or (
            not self.modn.startswith('fullsed')):
            self.proj, self.modn = _os.path.split(self.modn)
        else:
            self.proj = ''
        if not self.modn.startswith('mod'):
            self.modn = ''
        if len(self.modn) == 0:
            self.modn = 'mod' + self.suf.split('mod')[-1].split('_')[0]
        self.suf = '_'+'_'.join(self.suf.split('mod')[-1].split('_')[1:])
        if hasattr(self, 'vdict'):
            if _os.path.splitext(fname)[1] == '.log':
                self._fill()
            else:
                self._fillfname()
        else:
            _warn.warn('Warning! No `vdict` found for HdustMod')

    def _fill(self):
        f0 = open(self.fname).read().split('\n')
        for it in self.vdict.items():
            setattr(self, it[0], _phc.fltTxtOccur(it[1], f0))
        return

    def _fillfname(self):
        for it in self.vdict.keys():
            setattr(self, it, _phc.fltTxtOccur(it, self.suf))
        return

    def get_nob(self):
        self.nob = int(_hdt.sed2info(self.fname)[1])
        return self.nob

    def get_lum(self):
        if _os.path.splitext(self.fname)[1] == '.log':
            logname = self.fname
        else:
            bdir = self.proj
            if bdir == _os.path.split(_os.getcwd())[1]:
                bdir = ""
            logname = _os.path.join(bdir, self.modn, self.modn + self.suf + 
                '.log')
            if not _os.path.exists(logname):
                loglist = _glob(_os.path.join(bdir, self.modn, '*'+self.modn + 
                    self.suf + '*.log'))
                logname = [l for l in loglist if _os.path.getsize(l) > 500]
                if len(logname) == 0:
                    raise LookupError('Not found an log file to {0}'.format(
                        self.fname))
                logname = logname[0]
        #
        self.lum = _phc.fltTxtOccur('L =', open(logname).read().split('\n'))
        if _np.isnan(self.lum):
            raise ValueError('# ERROR! Lum is NaN. Check '+logname)
        self.vrot = _phc.fltTxtOccur('Vrot ', open(logname).read().split('\n'))
        self.log = logname
        return self.lum

    def get_sig0(self):
        self.get_lum()
        self.tf = _phc.fltTxtOccur('Fraction ', open(self.log).read().split(
            '\n'))/100.
        self.tp = _phc.fltTxtOccur('Teff_pole ', open(self.log).read().split(
            '\n'))
        self.rp = _phc.fltTxtOccur(' R_pole', open(self.log).read().split(
            '\n')) 
        self.rf = _phc.fltTxtOccur('R_eq/R_pole ', open(self.log).read().split(
            '\n')) 
        self.mu = _phc.fltTxtOccur('mu =', open(self.log).read().split('\n')) 
        self.M = _phc.fltTxtOccur(' M =', open(self.log).read().split('\n')) 
        self.n0 = _phc.fltTxtOccur('n_0', open(self.log).read().split('\n'))
        self.sig0 = _hdt.n0toSigma0(self.n0, self.M, self.rp*self.rf, 
            self.tf/100., self.tp, self.mu)

    def readfs2ob(self, pol=False):
        if not hasattr(self, 'nob'):
            self.get_nob()
        col = 3
        if pol:
            col = 7
        s2d = _hdt.readfullsed2(self.fname)
        self.arr_lbd = s2d[0, :, 2]
        self.arr_flx = s2d[:, :, col]
        self.obdeg = _np.round(_np.arccos(s2d[:, 0, 0])*180./_np.pi, 1)
        self.obs = s2d[:, 0, 0]
        return

    def __str__(self):
        if self.fname is None:
            return 'None'
        return '_'+'_'.join(it[0]+it[1].format(getattr(self, it[0])) for it in 
            self.vdict.items())


class AeriMod(HdustMod):    
    """docstring for AeriMod"""
    vdict = _OrderedDict(zip(['dval', 'ht', "nr", 'renv', 'M', 'vt'],
            ['n_0', 'Fraction', ' n ', 'R_env', ' M ', 'V_turb'])) 
    vfmt = _OrderedDict(zip(['dval', 'ht', "nr", 'renv', 'M', 'vt'], 
        ['{:.1e}', '{:03.0f}', '{:.1f}', '{:04.1f}', '{:04.1f}', '{:03.0f}']))

    def __init__(self, fname):
        super(AeriMod, self).__init__(fname)


class PhotMod(HdustMod):    
    """docstring for AeriMod"""
    vdict = _OrderedDict(zip(['W', 'M'], ['w =', ' M '])) 
    vfmt = _OrderedDict(zip(['W', 'M'], ['{:.3f}', '{:04.1f}']))

    def __init__(self, fname):
        super(PhotMod, self).__init__(fname)


def makeDiskGrid(modn='01', mhvals=[1.5], hvals=[60.], rdvals=[18.6], 
    mvals=[2.0], sig0vals=[1.], convSig2Rho=False, doFVDD=False, sBdays=None, 
    sBfiles=None, selsources='*', alpha=.5, mu=.5, R0r=300, Mdot11=False, 
    path=None):
    """
    | ###CONFIG. OPTIONS
    | #MODEL NUMBER
    | modn        = '02'
    | #The following filter will be applied to the SOURCE selection (string 
    | # fmt)
    | selsources = '*'
    |                              
    | #SUPERFICIAL DENSITY PROFILE EXPONENT
    | mvals       = [1.5,2.0,2.5,3.0]
    | #VERTICAL DENSITY PROFILE EXPONENT     
    | mhvals      = [1.5]
    | #FRACTION OF TEFF OF PRIMARY STAR
    | #This parameter sets if it you be FIXED to OB=1.1 case
    | hvals       = [72.]
    | #DISK RADIUS EQUATORIAL...    
    | rdvals      = [30.]
    | #SIGMA_0 VALUES
    | sig0vals    = _np.logspace(_np.log10(0.02),_np.log10(4.0),7)
    | 
    | #Do the Full VDD model for the corresponding sig0?
    | doFVDD = True 
    | alpha = 0.5
    | mu = 0.5
    | #WARNING: it only generates a single R0 value per batch. If you want to 
    | # change it, run it twice (or more)
    | R0r = 300
    | ###END CONFIG.
    |
    | Mdot11 = TODO
    |
    | WARNING: if convSig2Rho=True, the routine assumes ``sig0vals`` contains
    |  rho0vals. For that, (sBfiles==None or sBdays==None) and doFVDD==False.
    """
    # Consistency check
    if convSig2Rho and not ((sBfiles is None or sBdays is None) and doFVDD is 
        False):
        print('# ERROR! There is an inconsistency in the input.')
        print('# Check the *convSig2Rho*!')
    # 
    G = _phc.G.cgs
    Msun = _phc.Msun.cgs
    Rsun = _phc.Rsun.cgs
    kB = _phc.kB.cgs
    mH = _phc.mH.cgs
    yr = _phc.yr.cgs

    def doPL(prodI):
        '''
        Given a prodI (i.e., src,sig0,rd,h,m,mh), generates the Power-Law model 
        input
        '''
        src, sig0, rd, h, m, mh = prodI
        M, Req, Tp = _rot.readscr(src)
        Mstr = str(M)
        M *= Msun
        Req *= Rsun

        Th = h * Tp / 100.
        # a0 = (kB*h/100.*Tp/mu/mH)**.5
        a = (kB * Th / mu / mH)**.5
        n0 = (G * M / 2. / _np.pi)**.5 * sig0 / mu / mH / a / Req**1.5
        # Th = a**2*mu*mH/kB

        srcname = src.replace('source/', '').replace('.txt', '')

        wmod = mod[:]
        wmod[13] = wmod[13].replace('18.6', ('%.2f' % rd))
        wmod[20] = wmod[20].replace('2.0', ('%.2f' % m))
        wmod[33] = wmod[33].replace('1.5', ('%.2f' % mh))
        wmod[40] = wmod[40].replace('18000.', ('%.1f' % Th))      
        if convSig2Rho:
            wmod[52] = wmod[52].replace('2.35E13', ('%.2e' % sig0))    
            suffix = '_PLn{0:.1f}_rho{1:5.2e}_h{2:03.0f}_Rd{3:05.1f}_{4}'.\
                format((m + mh), sig0, h, rd, srcname)  
        else:
            wmod[52] = wmod[52].replace('2.35E13', ('%.2e' % n0))    
            suffix = '_PLn{0:.1f}_sig{1:.2f}_h{2:03.0f}_Rd{3:05.1f}_{4}'.\
                format((m + mh), sig0, h, rd, srcname)  

        f0 = open('mod' + modn + '/mod' + modn + suffix + '.txt', 'w')
        f0.writelines(wmod)
        f0.close()
        return

    def doMdot(prodI):
        '''
        Given a prodI (i.e., src,sig0,rd,h,m,mh), generates the full VDD model 
        input
        '''
        src, sig0, rd, h, m, mh = prodI
        M, Req, Tp = _rot.readscr(src)
        Mstr = str(M)
        M *= Msun
        Req *= Rsun

        Th = h * Tp / 100.
        a = (kB * Th / mu / mH)**.5
        # a0 = (kB*h/100*Tp/mu/mH)**.5
        # a = a0*Req0*Req**.25/Req/Req**.25

        R0 = R0r * Req
        Mdot = sig0 * Req**2 * 3 * _np.pi * alpha * \
            a**2 / (G * M * R0)**.5  # SI units
        Mdot = Mdot / Msun * yr
        # Th = a**2*mu*mH/kB

        srcname = src.replace('source/', '').replace('.txt', '')
        # suffix = '_NI_Mdot{:.1e}_Rd{:.1f}_R0{:.1f}_alp{:.1f}_h{:.1f}_{}'.\
        # format(Mdot,rd,R0/Req,alpha,h,srcname)
        suffix = '_NIa{0:.1f}_sig{1:.2f}_h{2:03.0f}_Rd{3:05.1f}_{4}'.format(
            alpha, sig0, h, rd, srcname)

        wmod = mod[:]
        wmod[13] = wmod[13].replace('18.6', ('%.2f' % rd))
        wmod[18] = wmod[18].replace('1', ('%d' % 2))
        wmod[23] = wmod[23].replace('1.', ('%.2f' % alpha)) 
        wmod[24] = wmod[24].replace('= 0.', ('= %.2f' % (R0 / Req)))
        wmod[25] = wmod[25].replace('= 0', ('= %d' % 1))
        wmod[31] = wmod[31].replace('0', ('%d' % 1))  
        wmod[40] = wmod[40].replace('18000.', '{0:.1f}'.format(Th)) 
        wmod[49] = wmod[49].replace('2', ('%d' % 3))            
        wmod[55] = wmod[55].replace('1.E-9', ('%.2e' % Mdot))

        f0 = open('mod' + modn + '/mod' + modn + suffix + '.txt', 'w')
        f0.writelines(wmod)
        f0.close()
        return

    def doSB(prodI, hseq=False):
        '''
        Given a prodI (i.e., sources,rdvals,hvals,mhvals,sBdays,sBfiles),
        generates the Single Be based model input
        '''

        src, rd, h, mh, day, sfile = prodI
        M, Req, Tp = _rot.readscr(src)
        Mstr = str(M)
        M *= Msun
        Req *= Rsun

        Th = h * Tp / 100.
        # a0 = (kB*h/100.*Tp/mu/mH)**.5
        a = (kB * Th / mu / mH)**.5
        # n0 = (G*M/2./_np.pi)**.5*sig0/mu/mH/a/Req**1.5
        # Th = a**2*mu*mH/kB

        srcname = src.replace('source/', '').replace('.txt', '')

        wmod = mod[:]
        wmod[13] = wmod[13].replace('18.6', ('%.2f' % rd))
        wmod[18] = wmod[18].replace('= 1', '= 4')
        wmod[28] = wmod[28].replace(
            'deltasco/Atsuo/1D/data/dSco_a035_01', (sfile))
        wmod[29] = wmod[29].replace('2.3', ('%.2f' % (day / 365.25)))
        if not hseq:
            wmod[33] = wmod[33].replace('1.5', ('%.2f' % mh))
            suffix = '_SB{0}_{1:.1f}d_h{2:03.0f}_Rd{3:05.1f}_{4}'.format(
                _phc.trimpathname(sfile)[1], day, h, rd, srcname)
        else:
            wmod[31] = wmod[31].replace('= 0', '= 1')
            wmod[36] = wmod[36].replace('1.5', ('%.2f' % mh))
            suffix = '_SB{0}_{1:.1f}d_hseq_Rd{2:05.1f}_{3}'.format(
                _phc.trimpathname(sfile)[1], day, rd, srcname)

        wmod[40] = wmod[40].replace('18000.', ('%.1f' % Th))      

        f0 = open('mod' + modn + '/mod' + modn + suffix + '.txt', 'w')
        f0.writelines(wmod)
        f0.close()
        return

    # TODO Setup Tpole = REF of a (scale height)    
    # Tps = dict(zip(Ms, Tp11))

    # PROGRAM BEGINS
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
        if path[-1] != '/':
            path += '/'
    else:
        path = ''
    # Check modN folder
    if not _os.path.exists('mod{0}'.format(modn)):
        _os.system('mkdir mod{0}'.format(modn))

    # Select sources
    sources = _glob('source/' + selsources)

    # Load disk model
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_disco.txt'))
    mod = f0.readlines()
    f0.close()

    if sBdays is None or sBfiles is None:
        for prodI in _product(sources, sig0vals, rdvals, hvals, mvals, mhvals):
            doPL(prodI)
            i = 0
            if doFVDD:
                i = 1
                doMdot(prodI)
        print('# {0:.0f} arquivos foram gerados !!'.format(len(sources) *
            len(sig0vals) * len(rdvals) * len(hvals) * (len(mvals) + i) * 
            len(mhvals)))
    else:
        for prodI in _product(sources, rdvals, hvals, mhvals, sBdays, sBfiles):
            doSB(prodI)
            i = 0
            if doFVDD:
                i = 1
                doSB(prodI, hseq=True)
        print('# {0:.0f} arquivos foram gerados !!'.format(len(sources) *
            len(rdvals) * len(hvals) * len(sBdays) * (len(mhvals) + i) * 
            len(sBfiles)))

    if path is not '':
        _os.chdir(path0)
    # END PROGRAM    
    return


def makeInpJob(modn='01', nodes=512, simulations=['SED'],
    docases=[1, 3], sim1='step1', sim2='step1_ref', composition='pureH',
    controls='controls', gridcells='grid', observers='observers',
    images=[''], perturbations=None, clusters=['job'], srcid='',
    walltime='24:00:00', wcheck=False, email='$USER@localhost', chkout=False,
    st1max=20, st1refmax=24, ctrM=False, touch=False, srcNf=None, path=None):
    """
    Create INP+JOB files to run Hdust.

    All SOURCE files must initiate by "Be_". Otherwise, the `makeInpJob` will
    not work. This is to satisfies the criteria of a specific disk model for
    each source star.

        ### Start edit here ###
        modn = '02'

        #clusters config
        # job = AlphaCrucis; oar = MesoCentre Licallo; ge = MesoCentre FRIPP
        clusters = ['job','oar','ge','bgq']
        clusters = ['oar']
        nodes    = 48
        #if wcheck == True, walltime will be AUTOMATICALLY estimated
        walltime = '3:00:00'
        wcheck   = True
        email    = 'user@gmail.com'

        #Check if the outputs already exist
        chkout = True
        #Above the values below, the step1 will be considered done!
        st1max = 26
        st1refmax = 30
        #Gera inp+job so' para o source com '1.45' no nome
        #Nao funciona caracteres especiais como * ou ?
        srcid       = '1.45'
        srcid       = ''
        #Se um dos 3 casos nao estiver presente, ele gera input comentado.
        docases = [1,2,3]
        #1 = step1  <> Gera inp+job so' para mod#/mod#.txt (SEM source, so disco)
        #habilita ADDSUFFIX; retira OBSERVERS e IMAGES
        sim1 = 'step1'
        #2 = step1_refine 
        sim2 = 'step1_refine'
        #3 = outros <> Gera inp+job so' para mod#/mod#SOURCE.txt (post-proc.)
        #retira ADDSUFFIX; adiciona OBSERVERS (e talvez IMAGES)
        simulations = ['sed','h','brg','halpha','uv']
        simulations = ['sed_sig','brg_M','halpha_M','uv','j','h','k','l','m','n',
         'q1','q2']
        simulations = ['SED','Ha']
        images      = ['','h','brg','halpha','uv']
        images        = simulations[:]
        composition = 'pureH'
        controls    = 'no_op'
        controls    = 'controls'
        ctrM        = False
        gridcells   = 'grid'
        observers   = 'obs'
        touch       = True
        ###stop edition here
    """
    def isFloat(x):
        try:
            a = float(x)
        except ValueError:
            return False
        else:
            return True

    def doCase1(inp, cases):
        case1 = inp[:]
        case1[0] = case1[0].replace('suffix', suf)
        case1[1] = case1[1].replace('pureH', composition)
        if ctrM:
            i = suf.find('_M')
            M = suf[i:i + 7]
            case1[2] = case1[2].replace('controls', controls + M)
        else:
            case1[2] = case1[2].replace('controls', controls)
        case1[3] = case1[3].replace('grid', gridcells)
        case1[4] = case1[4].replace('step1', sim1)
        case1[5] = case1[5].replace('source', src)
        if perturbations is not None:
            case1.append("PERTURBATIONS = '{0}'\n".format(perturbations))
        if 1 not in cases:
            for i in range(len(case1)):
                case1[i] = '! ' + case1[i]
        return case1

    def doCase2(inp, cases):
        case1 = inp[:]
        case1[0] = case1[0].replace('suffix', suf)
        case1[1] = case1[1].replace('pureH', composition)
        if ctrM:
            i = suf.find('_M')
            M = suf[i:i + 7]
            case1[2] = case1[2].replace('controls', controls + M)
        else:
            case1[2] = case1[2].replace('controls', controls)
        case1[3] = case1[3].replace('grid', gridcells)
        case1[4] = case1[4].replace('step1', sim2)
        case1[5] = case1[5].replace('source', src)
        if perturbations is not None:
            case1.append("PERTURBATIONS = '{0}'\n".format(perturbations))
        if 2 not in cases:
            for i in range(len(case1)):
                case1[i] = '! ' + case1[i]
        return case1

    def doCase3(inp, simchk):
        case3 = []
        for i in range(len(simulations)):
            case1 = inp[:]
            case1[0] = case1[0].replace('suffix', suf)
            case1[1] = case1[1].replace('pureH', composition)
            if ctrM:
                j = suf.find('_M')
                M = suf[j:j + 7]
                case1[2] = case1[2].replace('controls', controls + M)
            else:
                case1[2] = case1[2].replace('controls', controls)
            case1[3] = case1[3].replace('grid', gridcells)
            case1[5] = case1[5].replace('source', src)
            if simulations[i] == 'SED':
                sig = suf[suf.find('_sig') + 4:suf.find('_sig') + 8]
                if isFloat(sig) and srcNf[i]:
                    case1[4] = case1[4].replace(
                        'step1', 'SED_sig{0}'.format(sig))
                else:
                    case1[4] = case1[4].replace('step1', simulations[i])
            elif srcNf[i]:
                case1[4] = case1[4].replace('step1', '{0}_{1}'.format(
                    simulations[i], src))
            else:
                case1[4] = case1[4].replace('step1', simulations[i])
            case1.append("OBSERVERS   = '{0}'\n".format(observers))
            if images[i] != '':
                case1.append("IMAGES      = '{0}'\n".format(images[i]))
            if perturbations is not None:
                case1.append("PERTURBATIONS = '{0}'\n".format(perturbations))
            case1.append('\n')
            if not simchk[i]:
                for i in range(len(case1)):
                    case1[i] = '! ' + case1[i]
            case3 += case1
        return case3

    def doJobs(mod, sel, nodes, addtouch='\n'):
        # load Ref
        f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF.{0}'.format(sel)))
        wout = f0.readlines()
        f0.close()

        outname = mod[mod.find('/') + 1:].replace('txt', sel)

        f0 = open('{0}s/{0}s_{1}_mod{2}.sh'.format(sel, proj, modn), 'a')
        if sel == 'job':
            wout[4] = wout[4].replace('128', '{0}'.format(nodes))
            wout[4] = wout[4].replace('36:00:00', '{0}'.format(walltime))
            wout[8] = wout[8].replace(
                'alexcarciofi@gmail.com', '{0}'.format(email))
            wout[11] = wout[11].replace('hdust_bestar2.02.inp', '{0}/{1}'.
            format(proj, mod.replace('.txt', '.inp')))
            if touch:
                wout[24] = addtouch
                modchmod = _phc.trimpathname(mod)
                modchmod[1] = modchmod[1].replace('.txt', '*')
                wout[31] = ('chmod -f 664 {0}/{1}/*{2}\nchmod -f 664 log/*\n' 
                    'chmod -f 664 ../../tmp/*\n'.format(proj, *modchmod))
            f0.writelines('qsub {0}/{1}s/{2}\n'.format(proj, sel, outname))
        elif sel == 'oar':
            wout[2] = wout[2].replace('12', '{0}'.format(int(round(
                nodes/12.))))
            wout[2] = wout[2].replace('24:0:0', '{0}'.format(walltime))
            wout[10] = wout[10].replace('hdust_bestar2.02.inp', '{0}/{1}'.
            format(proj, mod.replace('.txt', '.inp')))
            f0.writelines(
                'chmod -f a+x {0}/{1}s/{2}\n'.format(proj, sel, outname))
            f0.writelines(
                'oarsub -S ./{0}/{1}s/{2}\n'.format(proj, sel, outname))      
        elif sel == 'ge':
            wout[3] = wout[3].replace('48', '{0}'.format(nodes))
            wout[4] = wout[4].replace('45:00:00', '{0}'.format(walltime))
            wout[7] = wout[7].replace('dmfaes@gmail.com', '{0}'.format(email))
            wout[11] = wout[11].replace('hdust_bestar2.02.inp', '{0}/{1}'.
            format(proj, mod.replace('.txt', '.inp')))
            f0.writelines(
                'qsub -P hdust {0}/{1}s/{2}\n'.format(proj, sel, outname))
        elif sel == 'bgq':
            wout[14] = wout[14].replace('512', '{0}'.format(nodes))
            nodes = int(nodes)
            if nodes % 512 != 0:
                nrsv = (nodes // 512 + 1) * 128
            else:
                nrsv = (nodes // 512) * 128
            wout[10] = wout[10].replace('128', '{0}'.format(nrsv))
            wout[5] = wout[5].replace('24:00:00', '{0}'.format(walltime))
            wout[14] = wout[14].replace('hdust_bestar2.02.inp', '{0}/{1}'.
            format(proj, mod.replace('.txt', '.inp')))
            f0.writelines('chmod -f +x {0}/{1}s/{2}\n'.format(proj, sel, 
                outname))
            f0.writelines(
                'llsubmit ./{0}/{1}s/{2}\n'.format(proj, sel, outname))
        f0.close()

        f0 = open('{0}s/{1}'.format(sel, outname), 'w')
        f0.writelines(wout)
        print('# Saved: {0}s/{1}'.format(sel, outname))
        f0.close()
        return

    # PROGRAM START
    if srcNf is None:
        srcNf = len(simulations) * [False]
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
        if path[-1] != '/':
            path += '/'
    else:
        path = ''
    # obtain the actual directory
    proj = _os.getcwd() 
    proj = proj[proj.rfind('/') + 1:]

    # Folder's checks
    for sel in clusters:
        if not _os.path.exists('{0}s'.format(sel)):
            _os.system('mkdir {0}s'.format(sel))
        elif _os.path.exists('{0}s/{0}s_{1}_mod{2}.sh'.format(sel, proj, 
            modn)):
            _os.system('rm {0}s/{0}s_{1}_mod{2}.sh'.format(sel, proj, modn))

    # list of mods
    mods = _glob('mod{0}/mod{0}*.txt'.format(modn))

    # load REF_inp
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_inp.txt'))
    inp = f0.readlines()
    f0.close()

    for mod in mods:
        # Write inps
        f0 = open(mod.replace('.txt', '.inp'), 'w')
        f0.writelines('PROJECT = {0}\nMODEL = {1}\n\n'.format(proj, modn))

        suf = mod[mod.find('_'):-4]
        if mod.find('Be_') > -1:
            src = mod[mod.find('Be_'):-4]
        else:
            src = _os.path.splitext(_os.path.basename(srcid))[0]
        # src = 'aeri_draft'
        # if src.find(srcid) == -1:
        #     print('# Script skipped by missing "Be" in source...')
        #     continue

        cases = docases[:]
        # Do the touch thing
        addtouch = '\n'
        addtouch += 'chmod -f 664 ../../tmp/*\nchmod -f 664 {0}/mod{1}/*\n'.\
            format(proj, modn)
        if touch and ( (1 in cases) or (2 in cases) ):
            addtouch += 'touch {0}/{1}\n'.format(proj,
                                                 mod.replace('.txt', '.log'))
        if touch and 3 in cases:
            for sim in simulations:
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt', 
                    '.chk')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt', 
                    '.err')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt',
                    '.log')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt',
                    '_SEI.chk')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt',
                    '_SEI.err')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                addtouch += 'touch {0}/{1}\n'.format(proj, mod.replace('.txt', 
                    '_SEI.log')).replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim))
                err90a = '{0}/{1}'.format(proj, mod.replace('.txt', '.err').
                    replace('mod{0}/mod{0}'.format(modn), 'mod{0}/{1}_mod{0}'.
                    format(modn, sim)))
                err90b = '{0}/{1}'.format(proj, mod.replace('.txt', 
                    '_SEI.err').replace('mod{0}/mod{0}'.format(modn), 
                    'mod{0}/{1}_mod{0}'.format(modn, sim)))
                addtouch += 'touch {0}\n'.format(err90a[:90])
                addtouch += 'touch {0}\n'.format(err90b[:90])
                addtouch += 'touch {0}\n'.format(
                    err90a[:90].replace(".err", ".chk").replace(".er", ".ch").
                    replace(".e", ".c"))
                addtouch += 'touch {0}\n'.format(
                    err90b[:90].replace(".err", ".chk").replace(".er", ".ch").
                    replace(".e", ".c"))
        modchmod = _phc.trimpathname(mod)
        modchmod[1] = modchmod[1].replace('.txt', '*')
        # addtouch += 'chmod 664 {0}/{1}/*{2}\n'.format(proj, *modchmod)

        # Set simulation check variable
        if 3 in cases:
            simchk = _np.ones(len(simulations), dtype=bool)
        else:    
            simchk = _np.zeros(len(simulations), dtype=bool)

        if _os.path.exists(mod.replace('.txt', '{0:02d}.temp'.format(st1max)))\
            and chkout and 1 in cases:
            cases.remove(1)
        case1 = doCase1(inp, cases)
        f0.writelines(case1 + ['\n'])

        if _os.path.exists(mod.replace('.txt', '{0:02d}.temp'.format(
            st1refmax))) and chkout and 2 in cases:
            cases.remove(2)    
        case2 = doCase2(inp, cases)
        f0.writelines(case2 + ['\n'])

        if chkout and 3 in cases:
            for i in range(len(simulations)):
                sim = simulations[i]
                if sim.find('Be_') > 0:
                    sim = sim[:sim.find('Be_')-1]
                outs2a = 'mod{0}/{1}_mod{0}{2}.sed2'.format(
                    modn, sim, suf)
                outs2b = 'mod{0}/{1}_mod{0}{2}_SEI.sed2'.format(
                    modn, sim, suf)
                if _os.path.exists(outs2a) or _os.path.exists(outs2b):
                    simchk[i] = False
            if True not in simchk:
                cases.remove(3)
        case3 = doCase3(inp, simchk)
        f0.writelines(case3)
        f0.close()

        # Def automatic walltime:
        if wcheck:
            h = 0 
            if 1 in cases:
                h += 1
            if 2 in cases:
                h += 1
            idx = _np.where(simchk is True)
            if len(idx[0]) > 0:
                extra = 0 + 4 * len(idx[0])
                h = h + extra * 48 / nodes
            walltime = '{0}:0:0'.format(h)

        # Del old jobs
        for sel in clusters:
            outname = mod[mod.find('/') + 1:].replace('txt', sel)
            if _os.path.exists('{0}s/{1}'.format(sel, outname)):
                _os.system('rm {0}s/{1}'.format(sel, outname))

        # Write jobs (if necessary)
        if len(cases) > 0:
            for sel in clusters:
                doJobs(mod, sel, nodes, addtouch)

    if path is not '':
        _os.chdir(path0)
    # PROGRAM END
    return


def makeNoDiskGrid(modn, selsources, path=None):
    """
    #Create a model list with random disk parameters ("noCS" in filename)

    INPUT:  modn = '01'; selsources = '*' (filter that is applied to the SOURCE
    selection).

    OUTPUT: Files written
    """

    def doNoCS(src):
        '''
        Given a src, generates the noCS model input
        '''
        srcname = src.replace('source/', '').replace('.txt', '')
        suffix = '_noCS_{0}'.format(srcname)

        wmod = mod[:]
        # Remove a disk does not work:
        # wmod[9]=wmod[9].replace('1','0')
        wmod[13] = wmod[13].replace('18.6', '2.0')

        f0 = open('mod' + modn + '/mod' + modn + suffix + '.txt', 'w')
        f0.writelines(wmod)
        f0.close()
        return

    # PROGRAM BEGINS
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
        if path[-1] != '/':
            path += '/'
    else:
        path = ''
    # Check modN folder
    if not _os.path.exists('mod{0}'.format(modn)):
        _os.system('mkdir mod{0}'.format(modn))

    # Select sources
    sources = _glob('source/' + selsources)

    # Load disk model
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_disco.txt'))
    mod = f0.readlines()
    f0.close()

    for prodI in _product(sources):
        prodI = prodI[0]
        doNoCS(prodI)
    print('# {0:.0f} arquivos foram gerados !!'.format(len(sources)))
    if path is not "":
        _os.chdir(path0)
    # END PROGRAM
    return


def makeSimulLine(basesims, Rs, hwidth, Ms, Obs, lsuffix, path='source/'):
    """
    | vrots = [[167.023,229.187,271.072,301.299,313.702],
    |          [177.998,244.636,290.596,324.272,338.298],
    |          [192.612,267.017,318.288,355.320,370.638],
    |          [202.059,281.667,335.158,373.716,389.782],
    |          [209.244,292.409,358.626,410.439,430.844],
    |          [214.407,297.661,357.799,402.628,420.683]]

    | vrots = [[259.759,354.834,417.792,464.549,483.847],
    |          [252.050,346.163,406.388,449.818,468.126],
    |          [245.127,336.834,399.983,448.076,467.806],
    |          [239.522,329.496,388.734,432.532,450.806],
    |          [234.301,321.139,379.297,423.241,441.122],
    |          [228.538,313.797,370.343,412.488,429.914],
    |          [219.126,299.656,354.547,395.821,413.008],
    |          [211.544,288.840,341.081,380.426,396.978],
    |          [203.438,279.328,328.666,365.697,380.660],
    |          [197.823,268.964,316.901,353.568,368.506],
    |          [192.620,262.688,308.208,341.963,356.410],
    |          [187.003,255.125,299.737,332.511,346.043]]
    |         
    | basesims = ['simulation/Brg.txt','simulation/Ha.txt']
    | Rs = [12000, 20000]
    | 
    | Ms = [4.00,5.00,7.00,9.00,12.00,15.00]
    | Ms = [14.6, 12.5, 10.8, 9.6, 8.6, 7.7, 6.4, 5.5, 4.8, 4.2, 3.8, 3.4]
    | Obs = [1.1,1.2,1.3,1.4,1.45]
    | suffix = 'H0.30_Z0.014_bE_Ell'
    """
    c = _phc.c.cgs

    for prodI in _product(Ms, Obs, basesims, lsuffix):
        M, Ob, basesim, suffix = prodI

        f0 = open(basesim)
        mod = f0.readlines()
        f0.close()

        srcid = 'Be_M{0:05.2f}_ob{1:.2f}_{2}.txt'.format(M, Ob, suffix)

        # i = Ms.index(M)
        # j = Obs.index(Ob)
        k = basesims.index(basesim)
        R = Rs[k]
        nmod = mod[:]
        vrot = _rot.vrot_scr(path+srcid)
        vel = '{0:.1f}'.format(hwidth + vrot)
        nmod[103] = nmod[103].replace('1020.', vel)
        n = str(int(round(2 * (hwidth + vrot) * R / c * 1e5)))
        # if len(_np.shape(vrots)) == 2:
        #     vel = '{0:.1f}'.format(hwidth + vrots[i][j])
        #     nmod[103] = nmod[103].replace('1020.', vel)
        #     n = str(int(round(2 * (hwidth + vrots[i][j]) * R / c * 1e5)))
        # else:
        #     vel = '{0:.1f}'.format(hwidth + vrots[0])
        #     nmod[103] = nmod[103].replace('1020.', vel)
        #     n = str(int(round(2 * (hwidth + vrots[0]) * R / c * 1e5)))
        # print(srcid, n)
        nmod[100] = nmod[100].replace('100', n)

        f0 = open(
            basesim.replace('.txt', '_{0}'.format(srcid)), 'w')
        f0.writelines(nmod)
        f0.close()
    return


def makeSourceGrid(masses, rps, lums, Ws, betas, path=None):
    """ Arbitrary values (no stellar model info), for the SELF-CONSISTENT RIGID 
    ROTATOR setup.

    OUTPUT: The combination of the values of all lists in the *path/source/* 
    folder.

    Example: 2 of each parameters results in 32 models (2^5)
    """
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
        if path[-1] != '/':
            path += '/'
    else:
        path = ''
    # 
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_estrela.txt'))
    mod = f0.readlines()
    f0.close()
    if not _os.path.exists('{0}source'.format(path)):
        _os.system('mkdir {0}source'.format(path))
    # 
    for prod in _product(masses, rps, lums, Ws, betas):
        MI, Raio, Lum, W, Beta = prod
        # REGISTRA VALORES
        wmod = mod[:]        
        wmod[3] = wmod[3].replace('10.3065', ('%.2f' % MI))
        wmod[4] = wmod[4].replace('5.38462', ('%.2f' % Raio))
        wmod[5] = wmod[5].replace('0.775', ('%.4f' % W))            
        wmod[6] = wmod[6].replace('7500.', ('%.2f' % Lum))
        wmod[7] = wmod[7].replace('0.25', ('%.5f' % Beta))
        # 
        suffix = '_M{0:05.2f}_W{1:.2f}_b{2:.2f}_rp{3:05.2f}_L{4:05.0f}_Ell'. \
            format(MI, W, Beta, Raio, Lum)
        f0 = open('{0}source/Be'.format(path) + suffix + '.txt', 'w')
        f0.writelines(wmod)
        f0.close()
    #
    if path is not "":
        _os.chdir(path0)    
    return 


def makeStarGrid(oblats, Hfs, path=None):
    """
    Based on GENEVA models.

    | INPUT: oblats = [1.1,1.2,1.3,1.4,1.45] (example)
    | Hfs = [0.3] (example)

    Masses list a Z value are inside `geneve_par.pro` file.
    """
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
        if path[-1] != '/':
            path += '/'
    else:
        path = ''
    if not _os.path.exists('stmodels'):
        _os.system('mkdir stmodels')
    try:
        runIDL = True 
        import pidly
    except ImportError:
        print('# This system do not have pIDLy installed...')
        runIDL = False

    if runIDL:
        key = _phc.user_input('# Do you want to run "geneve_par" (y/other):')
        if key != 'y':
            runIDL = False

    if runIDL:
        idl = pidly.IDL()
        propath = _os.path.join( _hdt.hdtpath(), 'refs' )
        idl('cd,"{0}"'.format(propath))
        idl('.r geneve_par')
        for ob in oblats:
            for H in Hfs:
                idl('geneve_par, {0}, {1}, /OBLAT,/makeeps'.format(ob, H))
                _os.system('mv {0}/geneve_lum.eps stmodels/geneve_lum_' +
                    '{1:.2f}_{2:.2f}.eps'.format(propath, ob, H))
                _os.system(
                    'mv {0}/geneve_rp.eps stmodels/geneve_rp_{1:.2f}_' +
                    '{2:.2f}.eps'.format(propath, ob, H))
                _os.system(
                    'mv {0}/geneve_par.txt stmodels/oblat{1}_h{2}.txt'.
                    format(propath, ob, H))
        idl.close()

    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_estrela.txt'))
    mod = f0.readlines()
    f0.close()

    if not _os.path.exists('source'):
        _os.system('mkdir source')

    for ob in oblats:                        
        for H in Hfs:  
            f0 = open('stmodels/oblat{0}_h{1}.txt'.format(ob, H))                     
            matriz = f0.readlines() 
            f0.close()
            Omega, W, Beta = map(float, matriz[1].split())

            m2 = []
            for i in range(4, len(matriz)):
                if len(matriz[i]) > 1:
                    m2 += [matriz[i].split()[1:]]
            matriz = _np.array(m2, dtype=float)
            M = matriz[:, 0]  # MASS (SOLAR MASSES)
            M = list(M)
            Rp = matriz[:, 1]  # POLAR RADIUS (SOLAR RADII)
            Rp = list(Rp)
            L = matriz[:, 2]  # LUMINOSITY (in solar lum.)
            L = list(L)
            Z = [0.014]  # METALLICITY(=Zsolar) (other options: 0.006, 0.002)

            print('Omega = ', Omega)
            print('W     = ', W)
            print('beta  = ', Beta)
            print('M     = ', M)
            print('Rp    = ', Rp)
            print('L     = ', L)
            print("%.0f arquivos gerados\n" % (len(M) * len(Hfs)))

            # DEFINE ALL INDEX
            for prodI in _product(M, Rp, L, Z):
                MI, RpI, LI, ZI = prodI
                a = M.index(MI)
                Raio = Rp[a]
                Lum = L[a]

                suffix = '_M{0:05.2f}_ob{1:.2f}_H{2:.2f}_Z{3}_bE_Ell'. \
                    format(MI, ob, H, ZI, Beta, RpI, LI)

                # REGISTRA VALORES
                wmod = mod[:]        
                wmod[3] = wmod[3].replace('10.3065', ('%.2f' % MI))
                wmod[4] = wmod[4].replace('5.38462', ('%.2f' % Raio))
                wmod[5] = wmod[5].replace('0.775', ('%.4f' % W))            
                wmod[6] = wmod[6].replace('7500.', ('%.2f' % Lum))
                wmod[7] = wmod[7].replace('0.25', ('%.5f' % Beta))

                f0 = open('source/Be' + suffix + '.txt', 'w')
                f0.writelines(wmod)
                f0.close()
    #
    if path is not "":
        _os.chdir(path0)
    return


def makeSimulDens(dbase, basesim):
    """
    Sets the SED simulations number of photos so that the signal/noise level
    is approximately constant at visible polarization.

    |dbase = _np.logspace(_np.log10(0.02),_np.log10(4.0),7)
    |basesim = 'simulation/sed.txt'
    """
    f0 = open(basesim)
    mod = f0.readlines()
    f0.close()
    # fact = 2.  Tempo execucao = d/1e13*fact
    # Nf0 = 500000000
    for d in dbase:
        srcid = 'sig{0:.2f}'.format(d)
        # alpha = .39794
        # beta =  13.87219
        alpha = 0.34588
        beta = 8.50927
        newd = int(10**(-alpha * _np.log10(d) + beta))
        print('{0}, N_f = {1:.2f}e+9'.format(srcid, newd / 1e9))
        nmod = mod[:]
        nmod[9] = nmod[9].replace('500000000', '{0}'.format(newd))
        f0 = open(basesim.replace('.txt', '_{0}.txt'.format(srcid)), 'w')
        f0.writelines(nmod)
        f0.close()
    # a = raw_input('asdads')
    return 


def makeCSGrid_bistabWind1Dust(modn='01', renv=[18.6], rcs=[5.], 
    dm_dOmega=[1.5e-5], v_0=[10.], v_inf=[1200.], beta_v=[2], A1=[49], 
    A2=[-0.7], A3=[0.], m_dth=[92.], grain_dust_ratio=[200.], grain_dens=[1.], 
    selsources='*', path=None):
    """ Great a CS Grid for HDUST for wind+dust shell (1) both in bi-stability

    Based on Carciofi+2010 and used in the B[e] grid between IAG/ON (Brazil).

    :param arg1: the first value
    :param arg2: the first value
    :type arg1: int, float,...
    :type arg2: int, float,...
    :returns: arg1/arg2 +arg3
    :rtype: int, float

    :Example:

    >>> import template
    >>> a = template.MainClass1()
    >>> a.function1(1,1,1)
    2

    .. note:: can be useful to emphasize important feature
    .. seealso:: :class:`MainClass2`
    .. warning:: arg2 must be non-zero.
    .. todo:: check that arg2 is non zero.
    """
    def dobistabWind1Dust(lpars):
        '''
        Given a proper list of parameters, do the thing...
        '''
        dm_d0, v0, vinf, beta, a1, a2, a3, mdth, gdratio, gdens, re, rd, src =\
            lpars

        srcname = _os.path.splitext(_os.path.basename(src))[0]
        # TODO
        suffix = ('_rd{0:02.0f}_a1+{1:02.0f}_m{2:03.0f}_md{3:.0e}_grh{4:04.1f}'
            '_gdr{5:03.0f}_v0+{6:02.0f}_vinf{7:04.0f}_a2+{8:04.1f}_b{9:03.1f}'
            '_a3+{10:03.1f}').format(rd, a1, mdth, dm_d0, gdens, gdratio, v0, 
            vinf, a2, beta, a3, srcname)

        wmod = mod[:]
        wmod = _phc.repl_fline_val(wmod, 13, '18.6', re)
        wmod = _phc.repl_fline_val(wmod, 14, '3.', rd)
        wmod = _phc.repl_fline_val(wmod, 19, '9.', a1)
        wmod = _phc.repl_fline_val(wmod, 20, '182.', mdth)
        wmod = _phc.repl_fline_val(wmod, 26, '39.', a1)
        wmod = _phc.repl_fline_val(wmod, 27, '182.', mdth)
        wmod = _phc.repl_fline_val(wmod, 35, '2.E-9', dm_d0)
        wmod = _phc.repl_fline_val(wmod, 30, '1.E-7', dm_d0)
        wmod = _phc.repl_fline_val(wmod, 31, '1.', gdens)
        wmod = _phc.repl_fline_val(wmod, 32, '200.', gdratio)
        wmod = _phc.repl_fline_val(wmod, 47, '10.', v0)
        wmod = _phc.repl_fline_val(wmod, 48, '400.', vinf)
        wmod = _phc.repl_fline_val(wmod, 49, '-0.7', a2)
        wmod = _phc.repl_fline_val(wmod, 50, '0.8', beta)
        wmod = _phc.repl_fline_val(wmod, 51, '2.75', a3)

        fmod = _os.path.join('mod'+modn, 'mod'+modn+suffix+'.txt')
        f0 = open(fmod, 'w')
        f0.writelines(wmod)
        f0.close()
        return

    # PROGRAM BEGINS
    path0 = _os.getcwd()
    if path is not None:
        _os.chdir(path)
    else:
        path = ''
    # Check modN folder
    if not _os.path.exists('mod{0}'.format(modn)):
        _os.system('mkdir mod{0}'.format(modn))

    # Select sources
    # print _os.path.join('source', selsources)
    sources = _glob(_os.path.join('source', selsources))

    # Load disk model
    f0 = open(_os.path.join(_hdt.hdtpath(), 'refs', 'REF_bistabWind1Dust.txt'))
    mod = f0.readlines()
    f0.close()

    for lpars in _product(dm_dOmega, v_0, v_inf, beta_v, A1, A2, A3, m_dth, 
        grain_dust_ratio, grain_dens, renv, rcs, sources):
        dobistabWind1Dust(lpars)

    if path is not "":
        _os.chdir(path0)
    # END PROGRAM
    return


def check_inp(mods=None, step1=["step1", 20], step1_ref=["step1_refine", 30],
    cust_sim=[["SED", "SED"]], ignore_comm=True):
    """ To be executed in the PROJECT folder. 

    ``mods`` = array of model folders. Ex: ["mod01", "mod02"]. ``None`` 
    do it for all.

    ``cust_sim`` = specifies a pair of *simulation-prefix* to be check (by 
    default, these are equal).

    ``ignore_comm`` = ignore already comments simulations inputs.

    WARNING: It assumes that *SUFFIX* inside `inp` file is consistent with 
    the filename. The function also removes the ``*.oar`` and ``*.job`` files.
    """
    if mods is None:
        mods = _glob("mod*")
    rule = (r"^(?![#!])[A-Za-z0-9'\"\-_+\. =]*SIMULATION.*?=.*?"
        "([A-Za-z0-9_-]+).*?")
    if ignore_comm is False:
        rule = (r"SIMULATION.*?=.*?([A-Za-z0-9_-]+).*?")
    for m in mods:
        inps = _glob(_os.path.join(m, 'mod*.inp'))
        for i in inps:
            modname = _os.path.split(i)[1]
            modname = _os.path.splitext(modname)[0]
            f0 = open(i).read().split("\n")
            out = _re.findall(rule, open(i).read(), flags=_re.M)
            for sim in out:
                if sim == step1[0]:
                    chk = sorted(_glob(_os.path.join(m, "{0}*[0-9].temp".
                        format(modname))))
                    if len(chk) > 0:
                        last = chk[-1][-7:-5]
                        if int(last) >= step1[1]:
                            for j in range(len(f0)):
                                if (_re.search(r"SIMULATION.*?=.*?" + sim + 
                                r"[\"' \t\n]", f0[j]) is not None):
                                    f0[j] = "! "+f0[j]
                                    if f0[j].upper().find("SUFFIX") > -1:
                                        break
                                    k = j
                                    while f0[k].upper().find("SUFFIX") == -1:
                                        k -= 1
                                        f0[k] = "! "+f0[k]
                                    k = j
                                    while f0[k+1].upper().find("SUFFIX") == -1:
                                        f0[k+1] = "! "+f0[k+1]
                                        k += 1
                                    break
                elif sim == step1_ref[0]:
                    chk = sorted(_glob(_os.path.join(m, "{0}*[0-9].temp".
                        format(modname))))
                    if len(chk) > 0:
                        last = chk[-1][-7:-5]
                        if int(last) >= step1[1]:
                            for j in range(len(f0)):
                                if (_re.search(r"SIMULATION.*?=.*?" + sim + 
                                r"[\"' \t\n]", f0[j]) is not None):
                                    f0[j] = "! "+f0[j]
                                    if f0[j].upper().find("SUFFIX") > -1:
                                        break
                                    k = j
                                    while f0[k].upper().find("SUFFIX") == -1:
                                        k -= 1
                                        f0[k] = "! "+f0[k]
                                    k = j
                                    while f0[k+1].upper().find("SUFFIX") == -1:
                                        f0[k+1] = "! "+f0[k+1]
                                        k += 1
                                        if k == len(f0)-1:
                                            break
                                    break
                else:
                    prefix = sim
                    for c in cust_sim:
                        if c[0] == sim:
                            prefix = c[1]
                    chk = _glob(_os.path.join(m, "{0}_{1}*.sed2".format(prefix,
                        modname)))
                    if len(chk) == 1:
                        for j in range(len(f0)):
                            if (_re.search(r"SIMULATION.*?=.*?" + sim + 
                            r"[\"' \t\n]", f0[j]) is not None):
                                f0[j] = "! "+f0[j]
                                # print(f0[0])
                                if f0[j].upper().find("SUFFIX") > -1:
                                    break
                                k = j
                                while f0[k].upper().find("SUFFIX") == -1:
                                    k -= 1
                                    f0[k] = "! "+f0[k]
                                k = j
                                while f0[k+1].upper().find("SUFFIX") == -1:
                                    f0[k+1] = "! "+f0[k+1]
                                    k += 1
                                    if k == len(f0)-1:
                                        break
                                break
                    elif len(chk) > 1:
                        _warn.warn("Multiple outputs for "+prefix+'_'+modname)
            f1 = open(i, "w")
            f1.writelines("\n".join(f0))
            f1.close()
            found = False
            for j in range(len(f0)):
                if f0[j].upper().find("SUFFIX") == 0:
                    found = True
                    break
            if not found:
                if _os.path.exists(_os.path.join("oars", modname+".oar")):
                    print("# Removed "+modname+".oar")
                    _os.remove(_os.path.join("oars", modname+".oar"))
                if _os.path.exists(_os.path.join("jobs", modname+".job")):
                    _os.remove(_os.path.join("jobs", modname+".job"))
    return 


# MAIN ###
if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Program to normalize spectra from OPERA Echelle format based on derivative 
variations.
"""
import os
from glob import glob
import numpy as np
import pyhdust.spectools as spt
from scipy import interpolate
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter

__version__ = "0.9"
__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


def plot_orders(ax, fwl, fflx, mode='o'):
    for i in range(len(fwl)):
        ax.plot(fwl[i], fflx[i], mode, ms=1)
    return ax


def do_norm(wl, flx):
    xout, yout = spt.max_func_pts(wl, flx, ws=0.01, avgbox=5)
    tck = interpolate.splrep(xout, yout)
    ynew = interpolate.splev(xout, tck, der=1)
    dmed = np.median( np.abs(ynew) )
    l = len(wl)
    # ConditionA: small derivative 
    conda = ( np.abs(ynew) < dmed )  
    # ConditionB: in the limits
    condb = ( wl[int(l*0.15)] > xout ) | ( xout > wl[int(l*(1-.15))] )
    # xmax = phc.find_nearest(yout, np.max(yout), idx=True)
    # xmax = xout[xmax]
    # xmax = wl[l/2]
    # ConditionC: xout < max(flx) AND ynew > 0
    # condc = ( xout <= xmax ) & ( ynew >= 0 )
    # ConditionD: xout > max(flx) AND ynew < 0
    # condd = ( xout >= xmax ) & ( ynew <= 0 )
    conde = ( ( xout > 658.45 ) | ( xout < 654.1 ) )
    # idx = np.where( (( conda & (condd | condc) ) | condb) & (conde) )  
    idx = np.where( (conda | condb) & (conde) )  
    # & (condd | condc))
    xout = xout[idx]
    yout = yout[idx]
    tck = interpolate.splrep(xout, yout, k=3)
    wl = wl[2:-2]
    flx = flx[2:-2]
    ynew = interpolate.splev(wl, tck, der=0)
    return wl, flx, ynew


path = '.'
opsps = glob('*.spc.gz')
print opsps

avgspecs = []
for sp in opsps:
    os.system('gunzip -c {0} > tmp.txt'.format(sp))
    spec = np.loadtxt('tmp.txt', skiprows=11)

    fig, ax = plt.subplots()
    ax.plot(spec[:, 5], spec[:, 10], label='Leg.')
    # ax.set_title('Title')
    # ax.legend()
    plt.savefig(sp.replace('.spc.gz', '_ori'))
    plt.close(fig)

    # i = 0
    onorm = []
    fwl = []
    oflx = []
    ocont = []
    orders = np.unique(spec[:, 0])
    for o in orders:
        idx = np.where(spec[:, 0] == o)
        wl = spec[:, 5][idx]
        flx = spec[:, 8][idx]
        try:
            wl, flx, ynew = do_norm(wl, flx)
            fwl.append(wl)
            onorm.append(flx/ynew)
            ocont.append(ynew)
            oflx.append(flx)        
        except:
            print('# Error in order {0}'.format(o))

    fig, ax = plt.subplots()
    ax = plot_orders(ax, fwl, oflx, 'o')
    ax = plot_orders(ax, fwl, ocont, '-')
    plt.savefig(sp.replace('.spc.gz', '.orders'))
    plt.close(fig)

    swl, sflx = spt.sum_ec(fwl, oflx)
    swl, scont = spt.sum_ec(fwl, ocont)
    swl, snorm = spt.sum_ec(fwl, onorm)

    avgspecs.append(sp.replace('.spc.gz', '.rv.fits'))
    nans, x = (np.isnan(scont), lambda z: z.nonzero()[0])
    sflx[nans] = 1.
    scont[nans] = 1.
    idx = np.where(scont <= 1e-3)
    scont[idx] = 1.
    sflx[idx] = 1.
    final = sflx/scont
    idx = np.where((final > 0.986) & (final < 1.014))
    pflx = final[idx]
    pwl = swl[idx]
    pflx = np.interp(swl, pwl, savgol_filter(pflx, 3, 1, mode='nearest'))
    final = final/pflx
    spt.writeFits(final, swl*10, 
        savename=sp.replace('.spc.gz', '.rv.fits'), quiet=False, 
        externhd='../../{0}'.format(sp.replace('.spc.gz', '.fits')) )

    fig, ax = plt.subplots()
    ax.plot(swl, final)
    ax.set_ylim([0, 3])
    plt.savefig(figname=sp.replace('.spc.gz', ''))
    plt.close(fig)

if os.path.exists('tmp.txt'):
    os.system('rm tmp.txt')

spt.averagespecs(avgspecs)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

""" Program to normalize a FITS (*.fit*) or a 2-columns text file (other 
extensions) spectrum based on non-parametric fitting.
"""

from glob import glob
import numpy as np
import os
import sys
import pyfits as pf
import pyhdust.spectools as spt
from pyhdust.tabulate import tabulate as tab
from argparse import ArgumentParser

__version__ = "0.93"
__author__ = "Daniel Moser"
__email__ = "dmfaes@gmail.com"


class MyParser(ArgumentParser): 
    def error(self, message):
        sys.stderr.write('# ERROR! %s\n' % message)
        self.print_help()
        sys.exit(2)

parser = MyParser(description=__doc__)
parser.add_argument('--version', action='version', 
    version='%(prog)s {0}'.format(__version__))
parser.add_argument("INPUT", help=("String to retrive the spectrum(a) "
    "filename(s) (wildcards accepted)"))
# parser.add_argument("-l", "--line", action="store", dest="line", 
#     help=("Wavelength of the reference line [default: %(default)s]"), 
#     type=float, default=6562.79)

args = parser.parse_args()


def norm_text(fname):
    f0 = np.loadtxt(fname)
    ny = spt.normalize_spec(f0[:, 0], f0[:, 1])
    outname = '{0}_norm{1}'.format(*os.path.splitext(fname))
    # f0 = np.savetxt(outname, np.column_stack((f0[:, 0], ny)))
    f1 = open(outname, 'w')
    f1.writelines( tab(np.column_stack((f0[:, 0], ny)), tablefmt='plain') )
    print('# {0} saved!'.format(outname))
    f1.close()
    return


def norm_spec(fname):
    imfits = pf.open(fname)
    hdr = imfits[0].header
    flux = imfits[0].data
    wl = np.arange(len(flux)) * hdr['CDELT1'] + hdr['CRVAL1']
    nflux = spt.normalize_spec(wl, flux.astype(float))
    outname = '{0}_norm{1}'.format(*os.path.splitext(fname))
    hdu = pf.PrimaryHDU(nflux)
    hdulist = pf.HDUList([hdu])
    hdulist[0].header = hdr
    hdu.writeto(outname, clobber=True)
    print('# {0} saved!'.format(outname))
    imfits.close()
    return

if __name__ == '__main__':

    linput = args.INPUT

    for specfile in glob(linput):
        if specfile.find('_norm.') > 0:
            print('# Warning! Remove "_norm" from {0} if you want to normalize'
                ' it!'.format(specfile))
            continue
        if specfile.find('.fit') > 0:
            norm_spec(specfile)
        else:
            norm_text(specfile)

    if len(glob(linput)) == 0:
        print('# ERROR! {0} files were not found!'.format(linput))



# -*- coding:utf-8 -*-

"""PyHdust auxiliary module: third-part module for reading/writing OIFITS 
files.

I made some modifications, but I keep the original README below.

This module is NOT related to the OIFITS Python module provided at
http://www.mrao.cam.ac.uk/research/OAS/oi_data/oifits.html
It is a (better) alternative.

To open an existing OIFITS file, use the oifits.open(filename)
function.  This will return an oifits object with the following
members (any of which can be empty dictionaries or numpy arrays):

   array: a dictionary of interferometric arrays, as defined by the
   OI_ARRAY tables.  The dictionary key is the name of the array
   (ARRNAME).

   target: a numpy array of targets, as defined by the rows of the
   OI_TARGET table.

   wavelength: a dictionary of wavelength tables (OI_WAVELENGTH).  The
   dictionary key is the name of the instrument/settings (INSNAME).

   vis, vis2 and t3: numpy arrays of objects containing all the
   measurement information.  Each list member corresponds to a row in
   an OI_VIS/OI_VIS2/OI_T3 table.

This module makes an ad-hoc, backwards-compatible change to the OIFITS
revision 1 standard originally described by Pauls et al., 2005, PASP,
117, 1255.  The OI_VIS and OI_VIS2 tables in OIFITS files produced by
this file contain two additional columns for the correlated flux,
CFLUX and CFLUXERR , which are arrays with a length corresponding to
the number of wavelength elements (just as VISAMP/VIS2DATA).

The main purpose of this module is to allow easy access to your OIFITS
data within Python, where you can then analyze it in any way you want.
As of version 0.3, the module can now be used to create OIFITS files
from scratch without serious pain.  Be warned, creating an array table
from scratch is probably like nailing jelly to a tree.  In a future
verison this will become easier.

The module also provides a simple mechanism for combining multiple
oifits objects, achieved by using the '+' operator on two oifits
objects: result = a + b.  The result can then be written to a file
using result.save(filename).

Many of the parameters and their meanings are not specifically
documented here.  However, the nomenclature mirrors that of the OIFITS
standard, so it is recommended to use this module with the PASP
reference above in hand.

Beginning with version 0.3, the OI_VIS/OI_VIS2/OI_T3 classes now use
masked arrays for convenience, where the mask is defined via the
'flag' member of these classes.  Beware of the following subtlety: as
before, the array data are accessed via (for example) OI_VIS.visamp;
however, OI_VIS.visamp is just a method which constructs (on the fly)
a masked array from OI_VIS._visamp, which is where the data are
actually stored.  This is done transparently, and the data can be
accessed and modified transparently via the "visamp" hidden attribute.
The same goes for correlated fluxes, differential/closure phases,
triple products, etc.  See the notes on the individual classes for a
list of all the "hidden" attributes.

Example of OIfits merge (same target):

.. code:: python

    import pyhdust.oifits as oifits
    oidata1 = oifits.open('file1.fits')
    oidata2 = oifits.open('file2.fits')
    oifits.matchtargetbyname = True
    merge = oidata1 + oidata2
    oimerge.save('merged.fits')

:co-author: Daniel Moser
:license: Copyright 2014 Paul Boley
"""
from __future__ import print_function
import datetime as _datetime
import copy as _copy
import numpy as _np
import numbers as _numbers
import warnings as _warn

try:
    import pyfits as _pyfits
except ImportError:
    _warn.warn('pyfits module not installed!!!')

__author__ = "Paul Boley"
__email__ = "boley@mpia-hd.mpg.de"

_mjdzero = _datetime.datetime(1858, 11, 17)

matchtargetbyname = False
matchstationbyname = False
refdate = _datetime.datetime(2000, 1, 1)


def _plurals(count):
    if count != 1:
        return 's'
    return ''


def _array_eq(a, b):
    "Test whether all the elements of two arrays are equal."

    try:
        return not (a != b).any()
    except:
        return not (a != b)


class _angpoint(float):

    "Convenience object for representing angles."

    def __init__(self, angle):
        self.angle = angle

    def __repr__(self):
        return '_angpoint(%s)' % self.angle.__repr__()

    def __str__(self):
        return "%g degrees" % (self.angle)

    def __eq__(self, other):
        return self.angle == other.angle

    def __ne__(self, other):
        return not self.__eq__(other)    

    def asdms(self):
        """Return the value as a string in dms format,
        e.g. +25:30:22.55.  Useful for declination."""
        angle = self.angle
        if angle < 0:
            negative = True
            angle *= -1.0
        else:
            negative = False
        degrees = _np.floor(angle)
        minutes = _np.floor((angle - degrees) * 60.0)
        seconds = (angle - degrees - minutes / 60.0) * 3600.0
        try:
            if negative:
                return "-%02d:%02d:%05.2f" % (degrees, minutes, seconds)
            else:
                return "+%02d:%02d:%05.2f" % (degrees, minutes, seconds)
        except TypeError:
            return self.__repr__()

    def ashms(self):
        """Return the value as a string in hms format,
        e.g. 5:12:17.21.  Useful for right ascension."""
        angle = self.angle * 24.0 / 360.0

        hours = _np.floor(angle)
        minutes = _np.floor((angle - hours) * 60.0)
        seconds = (angle - hours - minutes / 60.0) * 3600.0
        try:
            return "%02d:%02d:%05.2f" % (hours, minutes, seconds)
        except TypeError:
            return self.__repr__()


def getDate(hdu, hdulist):
    """ Get the header value of DATE-OBS from a hdu.
    If it is not valid, take it from "hdulist[0]".

    DEFINITION: The date of the observation, in the format specified in the
    FITS Standard.  The old date format was 'yy/mm/dd' and may be used only
    for dates from 1900 through 1999.  The new Y2K compliant date format is
    'yyyy-mm-dd' or 'yyyy-mm-ddTHH:MM:SS[.sss]'.
    """
    header = hdu.header
    if header['DATE-OBS'].find('-') > 0:
        date = header['DATE-OBS'].split('-')
    else:
        header0 = hdulist[0].header
        date = header0['DATE-OBS'].split('-')
    #
    if len(date[2]) > 2:
        date[2] = date[2][:2]
    #
    return date


class HDRINFO:

    def __init__(self, target, mjd, dateobs, datereduc):
        self.target = target
        self.mjd = mjd
        self.dateobs = dateobs
        self.datereduc = datereduc

    def returninfo(self):
        return self.target, self.mjd, self.dateobs, self.datereduc


class OI_TARGET:

    def __init__(self, target, raep0, decep0, equinox=2000.0, ra_err=0.0, 
        dec_err=0.0, sysvel=0.0, veltyp='TOPCENT', veldef='OPTICAL', pmra=0.0, 
        pmdec=0.0, pmra_err=0.0, pmdec_err=0.0, parallax=0.0, para_err=0.0, 
        spectyp='UNKNOWN'):
        self.target = target
        self.raep0 = _angpoint(raep0)
        self.decep0 = _angpoint(decep0)
        self.equinox = equinox
        self.ra_err = ra_err
        self.dec_err = dec_err
        self.sysvel = sysvel
        self.veltyp = veltyp
        self.veldef = veldef
        self.pmra = pmra
        self.pmdec = pmdec
        self.pmra_err = pmra_err
        self.pmdec_err = pmdec_err
        self.parallax = parallax
        self.para_err = para_err
        self.spectyp = spectyp

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.target != other.target) or
            (self.raep0 != other.raep0) or
            (self.decep0 != other.decep0) or
            (self.equinox != other.equinox) or
            (self.ra_err != other.ra_err) or
            (self.dec_err != other.dec_err) or
            (self.sysvel != other.sysvel) or
            (self.veltyp != other.veltyp) or
            (self.veldef != other.veldef) or
            (self.pmra != other.pmra) or
            (self.pmdec != other.pmdec) or
            (self.pmra_err != other.pmra_err) or
            (self.pmdec_err != other.pmdec_err) or
            (self.parallax != other.parallax) or
            (self.para_err != other.para_err) or
            (self.spectyp != other.spectyp))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%s: %s %s (%g)" % (self.target, self.raep0.ashms(), 
            self.decep0.asdms(), self.equinox)

    def info(self):
        print(str(self))


class OI_WAVELENGTH:

    def __init__(self, eff_wave, eff_band=None):
        self.eff_wave = _np.array(eff_wave, dtype=_np.double).reshape(-1)
        if eff_band is None:
            eff_band = _np.zeros_like(eff_wave)
        self.eff_band = _np.array(eff_band, dtype=_np.double).reshape(-1)

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (not _array_eq(self.eff_wave, other.eff_wave)) or
            (not _array_eq(self.eff_band, other.eff_band)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%d wavelength%s (%.3g-%.3g um)" % (len(self.eff_wave), 
            _plurals(len(self.eff_wave)), 1e6 * _np.min(self.eff_wave), 1e6 * 
            _np.max(self.eff_wave))

    def info(self):
        print(str(self))


class AMBER_SPECTRUM:

    """ Class for AMBER_SPECTRUM
    """

    def __init__(self, wavelength, interfspec, interfspecerr,
        spectrum, spectrumerr, eff_wave, eff_band=None):
        self.wavelength = wavelength
        self.eff_wave = _np.array(eff_wave, dtype=_np.double).reshape(-1)
        if eff_band is None:
            eff_band = _np.zeros_like(eff_wave)
        self.eff_band = _np.array(eff_band, dtype=_np.double).reshape(-1)
        self.interfspec = _np.array(interfspec, dtype=_np.double).reshape(-1)
        self.interfspecerr = _np.array(
            interfspecerr, dtype=_np.double).reshape(-1)
        self._spectrum = _np.array(spectrum, dtype=_np.double).reshape(-1)
        self._spectrumerr = _np.array(
            spectrumerr, dtype=_np.double).reshape(-1)       

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.wavelength != other.wavelength) or
            (not _array_eq(self.eff_wave, other.eff_wave)) or
            (not _array_eq(self.eff_wave, other.eff_wave)) or
            (not _array_eq(self.interfspec, other.interfspec)) or
            (not _array_eq(self.interfspecerr, other.interfspecerr)) or
            (not _array_eq(self.spectrum, other.spectrum)) or
            (not _array_eq(self.spectrumerr, other.spectrumerr)))

    def __getattr__(self, attrname):
        if attrname in ('spectrum', 'spectrumerr'):
            return self.__dict__['_' + attrname]
        else:
            raise AttributeError(attrname)

    def __setattr__(self, attrname, value):
        if attrname in ('spectrum', 'spectrumerr'):
            self.__dict__['_' + attrname] = value
        else:
            self.__dict__[attrname] = value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%d wavelength%s (%.3g-%.3g um)" % (len(self.eff_wave), 
            _plurals(len(self.eff_wave)), 1e6 * _np.min(self.eff_wave), 1e6 * 
            _np.max(self.eff_wave))

    def info(self):
        print(str(self))


class OI_VIS:

    """
    Class for storing visibility amplitude and differential phase data.
    To access the data, use the following hidden attributes:

    visamp, visamperr, visphi, visphierr, flag;
    and possibly cflux, cfluxerr.

    """

    def __init__(self, timeobs, int_time, visamp, visamperr, visphi, visphierr,
        flag, ucoord, vcoord, wavelength, target, array=None, station=(None, 
        None), cflux=None, cfluxerr=None):
        self.timeobs = timeobs
        self.array = array
        self.wavelength = wavelength
        self.target = target
        self.int_time = int_time
        self._visamp = _np.array(visamp, dtype=_np.double).reshape(-1)
        self._visamperr = _np.array(visamperr, dtype=_np.double).reshape(-1)
        self._visphi = _np.array(visphi, dtype=_np.double).reshape(-1)
        self._visphierr = _np.array(visphierr, dtype=_np.double).reshape(-1)
        if cflux is not None:
            self._cflux = _np.array(cflux, dtype=_np.double).reshape(-1)
        else:
            self._cflux = None
        if cfluxerr is not None:
            self._cfluxerr = _np.array(cfluxerr, dtype=_np.double).reshape(-1)
        else:
            self._cfluxerr = None
        self.flag = _np.array(flag, dtype=bool).reshape(-1)
        self.ucoord = ucoord
        self.vcoord = vcoord
        self.station = station

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.timeobs != other.timeobs) or
            (self.array != other.array) or
            (self.wavelength != other.wavelength) or
            (self.target != other.target) or
            (self.int_time != other.int_time) or
            (self.ucoord != other.ucoord) or
            (self.vcoord != other.vcoord) or
            (self.array != other.array) or
            (self.station != other.station) or
            (not _array_eq(self.visamp, other.visamp)) or
            (not _array_eq(self.visamperr, other.visamperr)) or
            (not _array_eq(self.visphi, other.visphi)) or
            (not _array_eq(self.visphierr, other.visphierr)) or
            (not _array_eq(self.flag, other.flag)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, attrname):
        if attrname in ('visamp', 'visamperr', 'visphi', 'visphierr'):
            return _np.ma.masked_array(self.__dict__['_' + attrname], 
                mask=self.flag)
        elif attrname in ('cflux', 'cfluxerr'):
            if self.__dict__['_' + attrname] is not None:
                return _np.ma.masked_array(self.__dict__['_' + attrname], 
                    mask=self.flag)
            else:
                return None
        else:
            raise AttributeError(attrname)

    def __setattr__(self, attrname, value):
        if attrname in ('visamp', 'visamperr', 'visphi', 'visphierr', 'cflux', 
            'cfluxerr'):
            self.__dict__['_' + attrname] = value
        else:
            self.__dict__[attrname] = value

    def __repr__(self):
        meanvis = _np.ma.mean(self.visamp)
        if self.station[0] and self.station[1]:
            baselinename = ' (' + \
                self.station[0].sta_name + self.station[1].sta_name + ')'
        else:
            baselinename = ''
        return ('%s %s%s: %d point%s (%d masked), B = %5.1f m, PA = %5.1f ' +
            'deg, <V> = %4.2g') % (self.target.target, self.timeobs.
                strftime('%F %T'), baselinename, len(self.visamp), 
                _plurals(len(self.visamp)), _np.sum(self.flag), 
                _np.sqrt(self.ucoord**2 + self.vcoord**2), 
                _np.arctan2(self.ucoord, self.vcoord) * 180.0 / _np.pi % 180.0, 
                meanvis)

    def info(self):
        print(str(self))


class OI_VIS2:

    """
    Class for storing squared visibility amplitude data.
    To access the data, use the following hidden attributes:

    vis2data, vis2err

    """

    def __init__(self, timeobs, int_time, vis2data, vis2err, flag, ucoord, 
        vcoord, wavelength, target, array=None, station=(None, None)):
        self.timeobs = timeobs
        self.array = array
        self.wavelength = wavelength
        self.target = target
        self.int_time = int_time
        self._vis2data = _np.array(vis2data, dtype=_np.double).reshape(-1)
        self._vis2err = _np.array(vis2err, dtype=_np.double).reshape(-1)
        self.flag = _np.array(flag, dtype=bool).reshape(-1)
        self.ucoord = ucoord
        self.vcoord = vcoord
        self.station = station

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.timeobs != other.timeobs) or
            (self.array != other.array) or
            (self.wavelength != other.wavelength) or
            (self.target != other.target) or
            (self.int_time != other.int_time) or
            (self.ucoord != other.ucoord) or
            (self.vcoord != other.vcoord) or
            (self.array != other.array) or
            (self.station != other.station) or
            (not _array_eq(self.vis2data, other.vis2data)) or
            (not _array_eq(self.vis2err, other.vis2err)) or
            (not _array_eq(self.flag, other.flag)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, attrname):
        if attrname in ('vis2data', 'vis2err'):
            return _np.ma.masked_array(self.__dict__['_' + attrname], 
                mask=self.flag)
        else:
            raise AttributeError(attrname)

    def __setattr__(self, attrname, value):
        if attrname in ('vis2data', 'vis2err'):
            self.__dict__['_' + attrname] = value
        else:
            self.__dict__[attrname] = value

    def __repr__(self):
        meanvis = _np.ma.mean(self.vis2data)
        if self.station[0] and self.station[1]:
            baselinename = ' (' + \
                self.station[0].sta_name + self.station[1].sta_name + ')'
        else:
            baselinename = ''
        return ('%s %s%s: %d point%s (%d masked), B = %5.1f m, PA = %5.1f ' +
            'deg, <V^2> = %4.2g') % (self.target.target, 
                self.timeobs.strftime('%F %T'), baselinename, 
                len(self.vis2data), _plurals(len(self.vis2data)),
                _np.sum(self.flag), _np.sqrt(self.ucoord**2 + self.vcoord**2), 
                _np.arctan2(self.ucoord, self.vcoord) * 180.0 / _np.pi % 180.0,
                meanvis)

    def info(self):
        print(str(self))


class OI_T3:

    """
    Class for storing triple product and closure phase data.
    To access the data, use the following hidden attributes:

    t3amp, t3amperr, t3phi, t3phierr

    """

    def __init__(self, timeobs, int_time, t3amp, t3amperr, t3phi, t3phierr, 
        flag, u1coord, v1coord, u2coord, v2coord, wavelength, target, 
        array=None, station=(None, None, None)):
        self.timeobs = timeobs
        self.array = array
        self.wavelength = wavelength
        self.target = target
        self.int_time = int_time
        self._t3amp = _np.array(t3amp, dtype=_np.double).reshape(-1)
        self._t3amperr = _np.array(t3amperr, dtype=_np.double).reshape(-1)
        self._t3phi = _np.array(t3phi, dtype=_np.double).reshape(-1)
        self._t3phierr = _np.array(t3phierr, dtype=_np.double).reshape(-1)
        self.flag = _np.array(flag, dtype=bool).reshape(-1)
        self.u1coord = u1coord
        self.v1coord = v1coord
        self.u2coord = u2coord
        self.v2coord = v2coord
        self.station = station

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.timeobs != other.timeobs) or
            (self.array != other.array) or
            (self.wavelength != other.wavelength) or
            (self.target != other.target) or
            (self.int_time != other.int_time) or
            (self.u1coord != other.u1coord) or
            (self.v1coord != other.v1coord) or
            (self.u2coord != other.u2coord) or
            (self.v2coord != other.v2coord) or
            (self.array != other.array) or
            (self.station != other.station) or
            (not _array_eq(self.t3amp, other.t3amp)) or
            (not _array_eq(self.t3amperr, other.t3amperr)) or
            (not _array_eq(self.t3phi, other.t3phi)) or
            (not _array_eq(self.t3phierr, other.t3phierr)) or
            (not _array_eq(self.flag, other.flag)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, attrname):
        if attrname in ('t3amp', 't3amperr', 't3phi', 't3phierr'):
            return _np.ma.masked_array(self.__dict__['_' + attrname], 
                mask=self.flag)
        else:
            raise AttributeError(attrname)

    def __setattr__(self, attrname, value):
        if attrname in ('vis2data', 'vis2err'):
            self.__dict__['_' + attrname] = value
        else:
            self.__dict__[attrname] = value

    def __repr__(self):
        meant3 = _np.mean(self.t3amp[_np.where(self.flag is False)])
        if self.station[0] and self.station[1] and self.station[2]:
            baselinename = ' (' + self.station[0].sta_name + self.station[
                1].sta_name + self.station[2].sta_name + ')'
        else:
            baselinename = ''
        return ("%s %s%s: %d point%s (%d masked), B = %5.1fm, %5.1fm, <T3> " +
            "= %4.2g") % (self.target.target, self.timeobs.strftime('%F %T'), 
                baselinename, len(self.t3amp), _plurals(len(self.t3amp)), 
                _np.sum(self.flag), _np.sqrt(self.u1coord**2 + 
                    self.v1coord**2), _np.sqrt(self.u2coord**2 + 
                    self.v2coord**2), meant3)

    def info(self):
        print(str(self))


class OI_STATION:

    """ This class corresponds to a single row (i.e. single
    station/telescope) of an OI_ARRAY table."""

    def __init__(self, tel_name=None, sta_name=None, diameter=None, 
        staxyz=[None, None, None]):
        self.tel_name = tel_name
        self.sta_name = sta_name
        self.diameter = diameter
        self.staxyz = staxyz

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.tel_name != other.tel_name) or
            (self.sta_name != other.sta_name) or
            (self.diameter != other.diameter) or
            (not _array_eq(self.staxyz, other.staxyz)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '%s/%s (%g m)' % (self.sta_name, self.tel_name, self.diameter)


class OI_ARRAY:

    """Contains all the data for a single OI_ARRAY table.  Note the
    hidden convenience attributes latitude, longitude, and altitude."""

    def __init__(self, frame, arrxyz, stations=()):
        self.frame = frame
        self.arrxyz = arrxyz
        self.station = _np.empty(0)
        for station in stations:
            tel_name, sta_name, sta_index, diameter, staxyz = station
            self.station = _np.append(self.station, OI_STATION(
                tel_name=tel_name, sta_name=sta_name, diameter=diameter,
                staxyz=staxyz))

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        equal = not (
            (self.frame != other.frame) or
            (not _array_eq(self.arrxyz, other.arrxyz)))

        if not equal:
            return False

        # If position appears to be the same, check that the stations
        # (and ordering) are also the same
        if (self.station != other.station).any():
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, attrname):
        if attrname == 'latitude':
            radius = _np.sqrt((self.arrxyz**2).sum())
            return _angpoint(_np.arcsin(self.arrxyz[2] / radius) * 180.0 / 
                _np.pi)
        elif attrname == 'longitude':
            radius = _np.sqrt((self.arrxyz**2).sum())
            xylen = _np.sqrt(self.arrxyz[0]**2 + self.arrxyz[1]**2)
            return _angpoint(_np.arcsin(self.arrxyz[1] / xylen) * 180.0 / 
                _np.pi)
        elif attrname == 'altitude':
            radius = _np.sqrt((self.arrxyz**2).sum())
            return radius - 6378100.0  
        else:
            raise AttributeError(attrname)

    def __repr__(self):
        return '%s %s %g m, %d station%s' % (self.latitude.asdms(), 
            self.longitude.asdms(), self.altitude, len(self.station), 
            _plurals(len(self.station)))

    def info(self, verbose=0):
        """Print the array's center coordinates.  If verbosity >= 1,
        print information about each station."""
        print(str(self))
        if verbose >= 1:
            for station in self.station:
                print("   %s" % str(station))

    def get_station_by_name(self, name):

        for station in self.station:
            if station.sta_name == name:
                return station

        raise LookupError('No such station %s' % name)


class oifits:

    def __init__(self):

        self.wavelength = {}
        self.amberspec = _np.empty(0)
        self.target = _np.empty(0)
        self.array = {}
        self.vis = _np.empty(0)
        self.vis2 = _np.empty(0)
        self.t3 = _np.empty(0)
        self.hdrinfo = {}

    def __radd__(self, other):
        return self + other

    def __add__(self, other):
        """Consistently combine two separate oifits objects.  Note
        that targets can be matched by name only (e.g. if coordinates
        differ) by setting oifits.matchtargetbyname to True.  The same
        goes for stations of the array (controlled by
        oifits.matchstationbyname)"""
        if isinstance(other, _numbers.Number):
            print('# Warning! Ignoring number sum with oifits!!!')
            return self 

        # Don't do anything if the two oifits objects are not CONSISTENT!
        if not self.isconsistent() or not other.isconsistent():
            print('oifits objects are not consistent, bailing.')
            return

        new = _copy.deepcopy(self)
        if len(other.wavelength):
            wavelengthmap = {}
            for key in other.wavelength.keys():
                if key not in new.wavelength.keys():
                    new.wavelength[key] = _copy.deepcopy(other.wavelength[key])
                elif new.wavelength[key] != other.wavelength[key]:
                    raise ValueError('Wavelength tables have the same key ' + 
                        'but differing contents.')
                wavelengthmap[id(other.wavelength[key])] = new.wavelength[key]

        if len(other.target):
            targetmap = {}
            for otarget in other.target:
                for ntarget in new.target:
                    if matchtargetbyname and ntarget.target == otarget.target:
                        targetmap[id(otarget)] = ntarget
                        break
                    elif ntarget == otarget:
                        targetmap[id(otarget)] = ntarget
                        break
                    elif ntarget.target == otarget.target:
                        print('Found a target with a matching name, but ' 
                            'some differences in the target specification. '
                            'Creating a new target. Set oifits.'
                            'matchtargetbyname to True to override this '
                            'behavior.')
                # If 'id(otarget)' is not in targetmap, then this is a new
                # target and should be added to the array of targets
                if id(otarget) not in targetmap.keys():
                    try:
                        newkey = new.target.keys()[-1] + 1
                    except:
                        newkey = 1
                    target = _copy.deepcopy(otarget)
                    new.target = _np.append(new.target, target)
                    targetmap[id(otarget)] = target

        if len(other.array):
            stationmap = {}
            arraymap = {}
            for key, otharray in other.array.iteritems():
                arraymap[id(otharray)] = key
                if key not in new.array.keys():
                    new.array[key] = _copy.deepcopy(other.array[key])
                # If arrays have the same name but seem to differ, try
                # to combine the two (by including the union of both
                # sets of stations)
                for othsta in other.array[key].station:
                    for newsta in new.array[key].station:
                        if newsta == othsta:
                            stationmap[id(othsta)] = newsta
                            break
                        elif matchstationbyname and newsta.sta_name == \
                            othsta.sta_name:
                            stationmap[id(othsta)] = newsta
                            break
                        elif newsta.sta_name == othsta.sta_name and not \
                            matchstationbyname:
                            raise ValueError('Stations have matching names ' +
                                'but conflicting data.')
                    # If 'id(othsta)' is not in the stationmap
                    # dictionary, then this is a new station and
                    # should be added to the current array
                    if id(othsta) not in stationmap.keys():
                        newsta = _copy.deepcopy(othsta)
                        new.array[key].station = _np.append(
                            new.array[key].station, newsta)
                        stationmap[id(othsta)] = newsta
                        # Make sure that staxyz of the new station is relative
                        # to the new array center
                        newsta.staxyz = othsta.staxyz - \
                            other.array[key].arrxyz + new.array[key].arrxyz

        for vis in other.vis:
            if vis not in new.vis:
                newvis = _copy.copy(vis)
                # The wavelength, target, array and station objects
                # should point to the appropriate objects inside the
                # 'new' structure
                newvis.wavelength = wavelengthmap[id(vis.wavelength)]
                newvis.target = targetmap[id(vis.target)]
                if (vis.array):
                    newvis.array = new.array[arraymap[id(vis.array)]]
                    newvis.station = [None, None]
                    newvis.station[0] = stationmap[id(vis.station[0])]
                    newvis.station[1] = stationmap[id(vis.station[1])]
                new.vis = _np.append(new.vis, newvis)

        for vis2 in other.vis2:
            if vis2 not in new.vis2:
                newvis2 = _copy.copy(vis2)
                # The wavelength, target, array and station objects
                # should point to the appropriate objects inside the
                # 'new' structure
                newvis2.wavelength = wavelengthmap[id(vis2.wavelength)]
                newvis2.target = targetmap[id(vis2.target)]
                if (vis2.array):
                    newvis2.array = new.array[arraymap[id(vis2.array)]]
                    newvis2.station = [None, None]
                    newvis2.station[0] = stationmap[id(vis2.station[0])]
                    newvis2.station[1] = stationmap[id(vis2.station[1])]
                new.vis2 = _np.append(new.vis2, newvis2)

        for t3 in other.t3:
            if t3 not in new.t3:
                newt3 = _copy.copy(t3)
                # The wavelength, target, array and station objects
                # should point to the appropriate objects inside the
                # 'new' structure
                newt3.wavelength = wavelengthmap[id(t3.wavelength)]
                newt3.target = targetmap[id(t3.target)]
                if (t3.array):
                    newt3.array = new.array[arraymap[id(t3.array)]]
                    newt3.station = [None, None, None]
                    newt3.station[0] = stationmap[id(t3.station[0])]
                    newt3.station[1] = stationmap[id(t3.station[1])]
                    newt3.station[2] = stationmap[id(t3.station[2])]
                new.t3 = _np.append(new.t3, newt3)

        return(new)

    def __eq__(self, other):

        if type(self) != type(other):
            return False

        return not (
            (self.wavelength != other.wavelength) or
            (self.amberspec != other.amberspec) or
            (self.target != other.target).any() or
            (self.array != other.array) or
            (self.vis != other.vis).any() or
            (self.vis2 != other.vis2).any() or
            (self.t3 != other.t3).any())

    def __ne__(self, other):
        return not self.__eq__(other)

    def isvalid(self):
        """Returns True of the oifits object is both consistent (as
        determined by isconsistent()) and conforms to the OIFITS
        standard (according to Pauls et al., 2005, PASP, 117, 1255)."""

        warnings = []
        errors = []
        if not self.isconsistent():
            errors.append('oifits object is not consistent')
        if not self.target.size:
            errors.append('No OI_TARGET data')
        if not self.wavelength:
            errors.append('No OI_WAVELENGTH data')
        if not self.amberspec:
            errors.append('No AMBER_SPECTRUM data')
        else:
            for wavelength in self.wavelength.values():
                if len(wavelength.eff_wave) != len(wavelength.eff_band):
                    errors.append("eff_wave and eff_band are of different " + 
                        "lengths for wavelength table")  # '%s'" % key)
        if (self.vis.size + self.vis2.size + self.t3.size == 0):
            errors.append(
                'Need to have atleast one measurement table (vis, vis2 or t3)')
        for vis in self.vis:
            nwave = len(vis.wavelength.eff_band)
            if (len(vis.visamp) != nwave) or (len(vis.visamperr) != nwave) or \
                (len(vis.visphi) != nwave) or (len(vis.visphierr) != nwave) or\
                    (len(vis.flag) != nwave):
                errors.append("Data size mismatch for visibility measurement" + 
                    "0x%x (wavelength table has a length of %d)" % (id(vis), 
                    nwave))
        for vis2 in self.vis2:
            nwave = len(vis2.wavelength.eff_band)
            if (len(vis2.vis2data) != nwave) or (len(vis2.vis2err) != nwave) \
                or (len(vis2.flag) != nwave):
                errors.append("Data size mismatch for visibility^2 " + 
                    "measurement 0x%x (wavelength table has a length of %d)" % 
                    (id(vis), nwave))                                  
        for t3 in self.t3:
            nwave = len(t3.wavelength.eff_band)
            if (len(t3.t3amp) != nwave) or (len(t3.t3amperr) != nwave) or \
                (len(t3.t3phi) != nwave) or (len(t3.t3phierr) != nwave) or \
                    (len(t3.flag) != nwave):
                errors.append("Data size mismatch for visibility measurement" + 
                    " 0x%x (wavelength table has a length of %d)" % (id(vis), 
                    nwave))

        if warnings:
            print("*** %d warning%s:" % (len(warnings), 
                _plurals(len(warnings))))
            for warning in warnings:
                print('  ' + warning)
        if errors:
            print("*** %d ERROR%s:" % (len(errors), 
                _plurals(len(errors)).upper()))
            for error in errors:
                print('  ' + error)

        return not (len(warnings) or len(errors))

    def isconsistent(self):
        """Returns True if the object is entirely self-contained,
        i.e. all cross-references to wavelength tables, arrays,
        stations etc. in the measurements refer to elements which are
        stored in the oifits object.  Note that an oifits object can
        be 'consistent' in this sense without being 'valid' as checked
        by isvalid()."""

        for vis in self.vis:
            if vis.array and (vis.array not in self.array.values()):
                print('A visibility measurement (0x%x) refers to an array '
                    'which is not inside the main oifits object.' % id(vis))
                return False
            if ((vis.station[0] and (vis.station[0] not in 
                vis.array.station)) or (vis.station[1] and (vis.station[1] 
                    not in vis.array.station))):
                print('A visibility measurement (0x%x) refers to a station'
                    ' which is not inside the main oifits object.' % id(vis))
                return False
            if vis.wavelength not in self.wavelength.values():
                print('A visibility measurement (0x%x) refers to a '
                    'wavelength table which is not inside the main oifits '
                    'object.' % id(vis))
                return False
            if vis.target not in self.target:
                print('A visibility measurement (0x%x) refers to a target '
                    'which is not inside the main oifits object.' % id(vis))
                return False

        for vis2 in self.vis2:
            if vis2.array and (vis2.array not in self.array.values()):
                print('A visibility^2 measurement (0x%x) refers to an array '
                    'which is not inside the main oifits object.' % id(vis2))
                return False
            if ((vis2.station[0] and (vis2.station[0] not in 
                vis2.array.station)) or (vis2.station[1] and (vis2.station[1] 
                not in vis2.array.station))):
                print('A visibility^2 measurement (0x%x) refers to a station '
                    'which is not inside the main oifits object.' % id(vis))
                return False
            if vis2.wavelength not in self.wavelength.values():
                print('A visibility^2 measurement (0x%x) refers to a '
                    'wavelength table which is not inside the main oifits '
                    'object.' % id(vis2))
                return False
            if vis2.target not in self.target:
                print('A visibility^2 measurement (0x%x) refers to a target '
                    'which is not inside the main oifits object.' % id(vis2))
                return False

        for t3 in self.t3:
            if t3.array and (t3.array not in self.array.values()):
                print('A closure phase measurement (0x%x) refers to an array '
                    'which is not inside the main oifits object.' % id(t3))
                return False
            if ((t3.station[0] and (t3.station[0] not in t3.array.station)) or
                (t3.station[1] and (t3.station[1] not in t3.array.station)) or
                (t3.station[2] and (t3.station[2] not in t3.array.station))):
                print('A closure phase measurement (0x%x) refers to a station '
                    'which is not inside the main oifits object.' % id(t3))
                return False
            if t3.wavelength not in self.wavelength.values():
                print('A closure phase measurement (0x%x) refers to a '
                    'wavelength table which is not inside the main oifits '
                    'object.' % id(t3))
                return False
            if t3.target not in self.target:
                print('A closure phase measurement (0x%x) refers to a target '
                    'which is not inside the main oifits object.' % id(t3))
                return False

        return True

    def info(self, recursive=True, verbose=0):
        """Print out a summary of the contents of the oifits object.
        Set recursive=True to obtain more specific information about
        each of the individual components, and verbose to an integer
        to increase the verbosity level."""

        if self.wavelength:
            wavelengths = 0
            if recursive:
                print("======================================================")
                print("SUMMARY OF WAVELENGTH TABLES")
                print("======================================================")
            for key in self.wavelength.keys():
                wavelengths += len(self.wavelength[key].eff_wave)
                if recursive:
                    print("'%s': %s" % (key, str(self.wavelength[key])))
            print("%d wavelength table%s with %d wavelength%s in total" % 
                (len(self.wavelength), _plurals(len(self.wavelength)), 
                    wavelengths, _plurals(wavelengths)))
        if self.target.size:
            if recursive:
                print("======================================================")
                print("SUMMARY OF TARGET TABLES")
                print("======================================================")
                for target in self.target:
                    target.info()
            print("%d target%s" % (len(self.target), 
                _plurals(len(self.target))))
        if self.array:
            stations = 0
            if recursive:
                print("======================================================")
                print("SUMMARY OF ARRAY TABLES")
                print("======================================================")
            for key in self.array.keys():
                if recursive:
                    print(key + ':')
                    self.array[key].info(verbose=verbose)
                stations += len(self.array[key].station)
            print("%d array%s with %d station%s" % (len(self.array), 
                _plurals(len(self.array)), stations, _plurals(stations)))
        if self.vis.size:
            if recursive:
                print("======================================================")
                print("SUMMARY OF VISIBILITY MEASUREMENTS")
                print("======================================================")
                for vis in self.vis:
                    vis.info()
            print("%d visibility measurement%s" % (len(self.vis), 
                _plurals(len(self.vis))))
        if self.vis2.size:
            if recursive:
                print("======================================================")
                print("SUMMARY OF VISIBILITY^2 MEASUREMENTS")
                print("======================================================")
                for vis2 in self.vis2:
                    vis2.info()
            print("%d visibility^2 measurement%s" % (len(self.vis2), 
                _plurals(len(self.vis2))))
        if self.t3.size:
            if recursive:
                print("======================================================")
                print("SUMMARY OF T3 MEASUREMENTS")
                print("======================================================")
                for t3 in self.t3:
                    t3.info()
            print("%d closure phase measurement%s" % (len(self.t3), 
                _plurals(len(self.t3))))

    def save(self, filename):
        """Write the contents of the oifits object to a file in OIFITS
        format."""

        if not self.isconsistent():
            print('oifits object is not consistent, refusing to go further')
            return

        hdulist = _pyfits.HDUList()
        hdu = _pyfits.PrimaryHDU()
        hdu.header.update('DATE', _datetime.datetime.now().strftime(
            format='%F'), comment='Creation date')
        hdu.header.add_comment('Written by a modified OIFITS Python module')  
            # version %s' % __version__)
        hdu.header.add_comment('http://j.mp/pyhdust')

        wavelengthmap = {}
        hdulist.append(hdu)
        for insname, wavelength in self.wavelength.iteritems():
            wavelengthmap[id(wavelength)] = insname
            hdu = _pyfits.new_table(_pyfits.ColDefs((
                _pyfits.Column(name='EFF_WAVE', format='1E', unit='METERS', 
                    array=wavelength.eff_wave),
                _pyfits.Column(name='EFF_BAND', format='1E', unit='METERS', 
                    array=wavelength.eff_band)
            )))
            hdu.header.update('EXTNAME', 'OI_WAVELENGTH')
            hdu.header.update(
                'OI_REVN', 1, 'Revision number of the table definition')
            hdu.header.update(
                'INSNAME', insname, 'Name of detector, for cross-referencing')
            hdulist.append(hdu)

        targetmap = {}
        if self.target.size:
            target_id = []
            target = []
            raep0 = []
            decep0 = []
            equinox = []
            ra_err = []
            dec_err = []
            sysvel = []
            veltyp = []
            veldef = []
            pmra = []
            pmdec = []
            pmra_err = []
            pmdec_err = []
            parallax = []
            para_err = []
            spectyp = []
            for i, targ in enumerate(self.target):
                key = i + 1
                targetmap[id(targ)] = key
                target_id.append(key)
                target.append(targ.target)
                raep0.append(targ.raep0)
                decep0.append(targ.decep0)
                equinox.append(targ.equinox)
                ra_err.append(targ.ra_err)
                dec_err.append(targ.dec_err)
                sysvel.append(targ.sysvel)
                veltyp.append(targ.veltyp)
                veldef.append(targ.veldef)
                pmra.append(targ.pmra)
                pmdec.append(targ.pmdec)
                pmra_err.append(targ.pmra_err)
                pmdec_err.append(targ.pmdec_err)
                parallax.append(targ.parallax)
                para_err.append(targ.para_err)
                spectyp.append(targ.spectyp)

            hdu = _pyfits.new_table(_pyfits.ColDefs((
                _pyfits.Column(name='TARGET_ID', format='1I', array=target_id),
                _pyfits.Column(name='TARGET', format='16A', array=target),
                _pyfits.Column(
                    name='RAEP0', format='1D', unit='DEGREES', array=raep0),
                _pyfits.Column(
                    name='DECEP0', format='1D', unit='DEGREES', array=decep0),
                _pyfits.Column(
                    name='EQUINOX', format='1E', unit='YEARS', array=equinox),
                _pyfits.Column(
                    name='RA_ERR', format='1D', unit='DEGREES', array=ra_err),
                _pyfits.Column(
                    name='DEC_ERR', format='1D', unit='DEGREES', 
                    array=dec_err),
                _pyfits.Column(
                    name='SYSVEL', format='1D', unit='M/S', array=sysvel),
                _pyfits.Column(name='VELTYP', format='A8', array=veltyp),
                _pyfits.Column(name='VELDEF', format='A8', array=veldef),
                _pyfits.Column(
                    name='PMRA', format='1D', unit='DEG/YR', array=pmra),
                _pyfits.Column(
                    name='PMDEC', format='1D', unit='DEG/YR', array=pmdec),
                _pyfits.Column(
                    name='PMRA_ERR', format='1D', unit='DEG/YR', 
                    array=pmra_err),
                _pyfits.Column(
                    name='PMDEC_ERR', format='1D', unit='DEG/YR', 
                    array=pmdec_err),
                _pyfits.Column(
                    name='PARALLAX', format='1E', unit='DEGREES', 
                    array=parallax),
                _pyfits.Column(
                    name='PARA_ERR', format='1E', unit='DEGREES', 
                    array=para_err),
                _pyfits.Column(name='SPECTYP', format='16A', array=spectyp)
            )))
            hdu.header.update('EXTNAME', 'OI_TARGET')
            hdu.header.update(
                'OI_REVN', 1, 'Revision number of the table definition')
            hdulist.append(hdu)

        arraymap = {}
        stationmap = {}
        for arrname, array in self.array.iteritems():
            arraymap[id(array)] = arrname
            tel_name = []
            sta_name = []
            sta_index = []
            diameter = []
            staxyz = []
            if array.station.size:
                for i, station in enumerate(array.station, 1):
                    stationmap[id(station)] = i
                    tel_name.append(station.tel_name)
                    sta_name.append(station.sta_name)
                    sta_index.append(i)
                    diameter.append(station.diameter)
                    staxyz.append(station.staxyz)
                hdu = _pyfits.new_table(_pyfits.ColDefs((
                    _pyfits.Column(
                        name='TEL_NAME', format='16A', array=tel_name),
                    _pyfits.Column(
                        name='STA_NAME', format='16A', array=sta_name),
                    _pyfits.Column(
                        name='STA_INDEX', format='1I', array=sta_index),
                    _pyfits.Column(
                        name='DIAMETER', unit='METERS', format='1E', 
                        array=diameter),
                    _pyfits.Column(
                        name='STAXYZ', unit='METERS', format='3D', 
                        array=staxyz)
                )))
            hdu.header.update('EXTNAME', 'OI_ARRAY')
            hdu.header.update(
                'OI_REVN', 1, 'Revision number of the table definition')
            hdu.header.update('ARRNAME', arrname, 
                comment='Array name, for cross-referencing')
            hdu.header.update('FRAME', array.frame, comment='Coordinate frame')
            hdu.header.update('ARRAYX', array.arrxyz[0], 
                comment='Array center x coordinate (m)')
            hdu.header.update('ARRAYY', array.arrxyz[1], 
                comment='Array center y coordinate (m)')
            hdu.header.update('ARRAYZ', array.arrxyz[2], 
                comment='Array center z coordinate (m)')
            hdulist.append(hdu)

        if self.vis.size:
            # The tables are grouped by ARRNAME and INSNAME -- all
            # observations which have the same ARRNAME and INSNAME are
            # put into a single FITS binary table.
            tables = {}
            for vis in self.vis:
                nwave = vis.wavelength.eff_wave.size
                if vis.array:
                    key = (arraymap[id(vis.array)], 
                        wavelengthmap[id(vis.wavelength)])
                else:
                    key = (None, wavelengthmap[id(vis.wavelength)])
                if key in tables.keys():
                    data = tables[key]
                else:
                    data = tables[key] = {'target_id': [], 'time': [], 
                    'mjd': [], 'int_time': [], 'visamp': [], 'visamperr': [], 
                        'visphi': [], 'visphierr': [], 'cflux': [], 
                        'cfluxerr': [], 'ucoord': [], 'vcoord': [], 
                        'sta_index': [], 'flag': []}
                data['target_id'].append(targetmap[id(vis.target)])
                if vis.timeobs:
                    time = vis.timeobs - refdate
                    data['time'].append(
                        time.days * 24.0 * 3600.0 + time.seconds)
                    mjd = (vis.timeobs - _mjdzero).days + \
                        (vis.timeobs - _mjdzero).seconds / 3600.0 / 24.0
                    data['mjd'].append(mjd)
                else:
                    data['time'].append(None)
                    data['mjd'].append(None)
                data['int_time'].append(vis.int_time)
                if nwave == 1:
                    data['visamp'].append(vis.visamp[0])
                    data['visamperr'].append(vis.visamperr[0])
                    data['visphi'].append(vis.visphi[0])
                    data['visphierr'].append(vis.visphierr[0])
                    data['flag'].append(vis.flag[0])
                    if vis.cflux is not None:
                        data['cflux'].append(vis.cflux[0])
                    else:
                        data['cflux'].append(None)
                    if vis.cfluxerr is not None:
                        data['cfluxerr'].append(vis.cfluxerr[0])
                    else:
                        data['cfluxerr'].append(None)
                else:
                    data['visamp'].append(vis.visamp)
                    data['visamperr'].append(vis.visamperr)
                    data['visphi'].append(vis.visphi)
                    data['visphierr'].append(vis.visphierr)
                    data['flag'].append(vis.flag)
                    if vis.cflux is not None:
                        data['cflux'].append(vis.cflux)
                    else:
                        cflux = _np.empty(nwave)
                        cflux[:] = None
                        data['cflux'].append(cflux)
                    if vis.cfluxerr is not None:
                        data['cfluxerr'].append(vis.cfluxerr)
                    else:
                        cfluxerr = _np.empty(nwave)
                        cfluxerr[:] = None
                        data['cfluxerr'].append(cfluxerr)
                data['ucoord'].append(vis.ucoord)
                data['vcoord'].append(vis.vcoord)
                if vis.station[0] and vis.station[1]:
                    data['sta_index'].append([stationmap[id(vis.station[0])], 
                        stationmap[id(vis.station[1])]])
                else:
                    data['sta_index'].append([-1, -1])
            for key in tables.keys():
                data = tables[key]
                nwave = self.wavelength[key[1]].eff_wave.size

                hdu = _pyfits.new_table(_pyfits.ColDefs([
                    _pyfits.Column(name='TARGET_ID', format='1I', 
                        array=data['target_id']),
                    _pyfits.Column(name='TIME', format='1D', unit='SECONDS', 
                        array=data['time']),
                    _pyfits.Column(name='MJD', unit='DAY', format='1D', 
                        array=data['mjd']),
                    _pyfits.Column(name='INT_TIME', format='1D', 
                        unit='SECONDS', array=data['int_time']),
                    _pyfits.Column(name='VISAMP', format='%dD' % nwave, 
                        array=data['visamp']),
                    _pyfits.Column(name='VISAMPERR', format='%dD' % nwave, 
                        array=data['visamperr']),
                    _pyfits.Column(name='VISPHI', unit='DEGREES', 
                        format='%dD' % nwave, array=data['visphi']),
                    _pyfits.Column(name='VISPHIERR', unit='DEGREES', 
                        format='%dD' % nwave, array=data['visphierr']),
                    _pyfits.Column(name='CFLUX', format='%dD' % 
                        nwave, array=data['cflux']),
                    _pyfits.Column(name='CFLUXERR', format='%dD' %
                        nwave, array=data['cfluxerr']),
                    _pyfits.Column(name='UCOORD', format='1D', unit='METERS', 
                        array=data['ucoord']),
                    _pyfits.Column(name='VCOORD', format='1D', unit='METERS', 
                        array=data['vcoord']),
                    _pyfits.Column(name='STA_INDEX', format='2I', 
                        array=data['sta_index'], null=-1),
                    _pyfits.Column(name='FLAG', format='%dL' % nwave)
                ]))

                # Setting the data of logical field via the
                # _pyfits.Column call above with length > 1 (eg
                # format='171L' above) seems to be broken, atleast as
                # of PyFITS 2.2.2
                hdu.data.field('FLAG').setfield(data['flag'], bool)
                hdu.header.update('EXTNAME', 'OI_VIS')
                hdu.header.update(
                    'OI_REVN', 1, 'Revision number of the table definition')
                hdu.header.update('DATE-OBS', refdate.strftime('%F'), 
                    comment='Zero-point for table (UTC)')
                if key[0]:
                    hdu.header.update(
                        'ARRNAME', key[0], 'Identifies corresponding OI_ARRAY')
                hdu.header.update('INSNAME', key[1], 
                    'Identifies corresponding OI_WAVELENGTH table')
                hdulist.append(hdu)

        if self.vis2.size:
            tables = {}
            for vis in self.vis2:
                nwave = vis.wavelength.eff_wave.size
                if vis.array:
                    key = (arraymap[id(vis.array)], 
                        wavelengthmap[id(vis.wavelength)])
                else:
                    key = (None, wavelengthmap[id(vis.wavelength)])
                if key in tables.keys():
                    data = tables[key]
                else:
                    data = tables[key] = {'target_id': [], 'time': [], 
                    'mjd': [], 'int_time': [], 'vis2data': [], 'vis2err': [], 
                        'ucoord': [], 'vcoord': [], 'sta_index': [], 
                        'flag': []}
                data['target_id'].append(targetmap[id(vis.target)])
                if vis.timeobs:
                    time = vis.timeobs - refdate
                    data['time'].append(
                        time.days * 24.0 * 3600.0 + time.seconds)
                    mjd = (vis.timeobs - _mjdzero).days + \
                        (vis.timeobs - _mjdzero).seconds / 3600.0 / 24.0
                    data['mjd'].append(mjd)
                else:
                    data['time'].append(None)
                    data['mjd'].append(None)
                data['int_time'].append(vis.int_time)
                if nwave == 1:
                    data['vis2data'].append(vis.vis2data[0])
                    data['vis2err'].append(vis.vis2err[0])
                    data['flag'].append(vis.flag[0])
                else:
                    data['vis2data'].append(vis.vis2data)
                    data['vis2err'].append(vis.vis2err)
                    data['flag'].append(vis.flag)
                data['ucoord'].append(vis.ucoord)
                data['vcoord'].append(vis.vcoord)
                if vis.station[0] and vis.station[1]:
                    data['sta_index'].append([stationmap[id(vis.station[0])], 
                        stationmap[id(vis.station[1])]])
                else:
                    data['sta_index'].append([-1, -1])
            for key in tables.keys():
                data = tables[key]
                nwave = self.wavelength[key[1]].eff_wave.size

                hdu = _pyfits.new_table(_pyfits.ColDefs([
                    _pyfits.Column(name='TARGET_ID', format='1I', 
                        array=data['target_id']),
                    _pyfits.Column(name='TIME', format='1D', unit='SECONDS', 
                        array=data['time']),
                    _pyfits.Column(name='MJD', format='1D', unit='DAY', 
                        array=data['mjd']),
                    _pyfits.Column(name='INT_TIME', format='1D', 
                        unit='SECONDS', array=data['int_time']),
                    _pyfits.Column(name='VIS2DATA', format='%dD' %
                        nwave, array=data['vis2data']),
                    _pyfits.Column(name='VIS2ERR', format='%dD' %
                        nwave, array=data['vis2err']),
                    _pyfits.Column(name='UCOORD', format='1D', unit='METERS', 
                        array=data['ucoord']),
                    _pyfits.Column(name='VCOORD', format='1D', unit='METERS', 
                        array=data['vcoord']),
                    _pyfits.Column(name='STA_INDEX', format='2I', 
                        array=data['sta_index'], null=-1),
                    _pyfits.Column(name='FLAG', format='%dL' % nwave, 
                        array=data['flag'])
                ]))
                # Setting the data of logical field via the
                # _pyfits.Column call above with length > 1 (eg
                # format='171L' above) seems to be broken, atleast as
                # of PyFITS 2.2.2
                hdu.data.field('FLAG').setfield(data['flag'], bool)
                hdu.header.update('EXTNAME', 'OI_VIS2')
                hdu.header.update('OI_REVN', 1, 
                    'Revision number of the table definition')
                hdu.header.update('DATE-OBS', refdate.strftime('%F'), 
                    comment='Zero-point for table (UTC)')
                if key[0]:
                    hdu.header.update('ARRNAME', key[0], 
                        'Identifies corresponding OI_ARRAY')
                hdu.header.update('INSNAME', key[1], 
                    'Identifies corresponding OI_WAVELENGTH table')
                hdulist.append(hdu)

        if self.t3.size:
            tables = {}
            for t3 in self.t3:
                nwave = t3.wavelength.eff_wave.size
                if t3.array:
                    key = (arraymap[id(t3.array)], 
                        wavelengthmap[id(t3.wavelength)])
                else:
                    key = (None, wavelengthmap[id(t3.wavelength)])
                if key in tables.keys():
                    data = tables[key]
                else:
                    data = tables[key] = {'target_id': [], 'time': [], 
                        'mjd': [], 'int_time': [], 't3amp': [], 't3amperr': [],
                        't3phi': [], 't3phierr': [], 'u1coord': [], 
                        'v1coord': [], 'u2coord': [], 'v2coord': [],
                        'sta_index': [], 'flag': []}
                data['target_id'].append(targetmap[id(t3.target)])
                if t3.timeobs:
                    time = t3.timeobs - refdate
                    data['time'].append(
                        time.days * 24.0 * 3600.0 + time.seconds)
                    mjd = (t3.timeobs - _mjdzero).days + \
                        (t3.timeobs - _mjdzero).seconds / 3600.0 / 24.0
                    data['mjd'].append(mjd)
                else:
                    data['time'].append(None)
                    data['mjd'].append(None)
                data['int_time'].append(t3.int_time)
                if nwave == 1:
                    data['t3amp'].append(t3.t3amp[0])
                    data['t3amperr'].append(t3.t3amperr[0])
                    data['t3phi'].append(t3.t3phi[0])
                    data['t3phierr'].append(t3.t3phierr[0])
                    data['flag'].append(t3.flag[0])
                else:
                    data['t3amp'].append(t3.t3amp)
                    data['t3amperr'].append(t3.t3amperr)
                    data['t3phi'].append(t3.t3phi)
                    data['t3phierr'].append(t3.t3phierr)
                    data['flag'].append(t3.flag)
                data['u1coord'].append(t3.u1coord)
                data['v1coord'].append(t3.v1coord)
                data['u2coord'].append(t3.u2coord)
                data['v2coord'].append(t3.v2coord)
                if t3.station[0] and t3.station[1] and t3.station[2]:
                    data['sta_index'].append([stationmap[id(t3.station[0])], 
                        stationmap[id(t3.station[1])], 
                        stationmap[id(t3.station[2])]])
                else:
                    data['sta_index'].append([-1, -1, -1])
            for key in tables.keys():
                data = tables[key]
                nwave = self.wavelength[key[1]].eff_wave.size

                hdu = _pyfits.new_table(_pyfits.ColDefs((
                    _pyfits.Column(
                        name='TARGET_ID', format='1I', array=data['target_id']),
                    _pyfits.Column(
                        name='TIME', format='1D', unit='SECONDS', array=data['time']),
                    _pyfits.Column(
                        name='MJD', format='1D', unit='DAY', array=data['mjd']),
                    _pyfits.Column(
                        name='INT_TIME', format='1D', unit='SECONDS', array=data['int_time']),
                    _pyfits.Column(name='T3AMP', format='%dD' %
                                   nwave, array=data['t3amp']),
                    _pyfits.Column(name='T3AMPERR', format='%dD' %
                                   nwave, array=data['t3amperr']),
                    _pyfits.Column(name='T3PHI', format='%dD' %
                                   nwave, unit='DEGREES', array=data['t3phi']),
                    _pyfits.Column(name='T3PHIERR', format='%dD' %
                                   nwave, unit='DEGREES', array=data['t3phierr']),
                    _pyfits.Column(
                        name='U1COORD', format='1D', unit='METERS', array=data['u1coord']),
                    _pyfits.Column(
                        name='V1COORD', format='1D', unit='METERS', array=data['v1coord']),
                    _pyfits.Column(
                        name='U2COORD', format='1D', unit='METERS', array=data['u2coord']),
                    _pyfits.Column(
                        name='V2COORD', format='1D', unit='METERS', array=data['v2coord']),
                    _pyfits.Column(
                        name='STA_INDEX', format='3I', array=data['sta_index'], null=-1),
                    _pyfits.Column(name='FLAG', format='%dL' %
                                   nwave, array=data['flag'])
                )))
                # Setting the data of logical field via the
                # _pyfits.Column call above with length > 1 (eg
                # format='171L' above) seems to be broken, atleast as
                # of PyFITS 2.2.2
                hdu.data.field('FLAG').setfield(data['flag'], bool)
                hdu.header.update('EXTNAME', 'OI_T3')
                hdu.header.update(
                    'OI_REVN', 1, 'Revision number of the table definition')
                hdu.header.update(
                    'DATE-OBS', refdate.strftime('%F'), 'Zero-point for table (UTC)')
                if key[0]:
                    hdu.header.update(
                        'ARRNAME', key[0], 'Identifies corresponding OI_ARRAY')
                hdu.header.update(
                    'INSNAME', key[1], 'Identifies corresponding OI_WAVELENGTH table')
                hdulist.append(hdu)

        hdulist.writeto(filename, clobber=True)


def open(filename, quiet=False):
    """Open an OIFITS file."""

    newobj = oifits()
    targetmap = {}
    sta_indices = {}

    if not quiet:
        print("Opening %s" % filename)
    hdulist = _pyfits.open(filename)
    # First get all the OI_TARGET, OI_WAVELENGTH and OI_ARRAY tables
    for hdu in hdulist:
        header = hdu.header
        data = hdu.data
        if hdu == hdulist[0]:
            hdrobj, hdrmjd, hdrobs, hdrdat = ('', '', '', '')
            if 'OBJECT' in header:
                hdrobj = header['OBJECT']
                if hdrobj.upper() == 'OBJECT':
                    if 'HIERARCH ESO OBS TARG NAME' in header:
                        hdrobj = header['HIERARCH ESO OBS TARG NAME']
            if 'MJD-OBS' in header:
                hdrmjd = header['MJD-OBS']
            if 'DATE-OBS' in header:
                hdrobs = header['DATE-OBS']
            if 'DATE' in header:
                hdrdat = header['DATE']
            newobj.hdrinfo = HDRINFO(hdrobj, hdrmjd, hdrobs, hdrdat)
        if hdu.name == 'OI_WAVELENGTH':
            if newobj.wavelength is None:
                newobj.wavelength = {}
            insname = header['INSNAME']
            newobj.wavelength[insname] = OI_WAVELENGTH(
                data.field('EFF_WAVE'), data.field('EFF_BAND'))
        elif hdu.name == 'OI_TARGET':
            for row in data:
                target_id = row['TARGET_ID']
                target = OI_TARGET(target=row['TARGET'], raep0=row['RAEP0'], decep0=row['DECEP0'],
                                   equinox=row['EQUINOX'], ra_err=row[
                                       'RA_ERR'], dec_err=row['DEC_ERR'],
                                   sysvel=row['SYSVEL'], veltyp=row[
                                       'VELTYP'], veldef=row['VELDEF'],
                                   pmra=row['PMRA'], pmdec=row[
                                       'PMDEC'], pmra_err=row['PMRA_ERR'],
                                   pmdec_err=row['PMDEC_ERR'], parallax=row[
                                       'PARALLAX'],
                                   para_err=row['PARA_ERR'], spectyp=row['SPECTYP'])
                newobj.target = _np.append(newobj.target, target)
                targetmap[target_id] = target
        elif hdu.name == 'OI_ARRAY':
            if newobj.array is None:
                newobj.array = {}
            arrname = header['ARRNAME']
            frame = header['FRAME']
            arrxyz = _np.array(
                [header['ARRAYX'], header['ARRAYY'], header['ARRAYZ']])
            newobj.array[arrname] = OI_ARRAY(frame, arrxyz, stations=data)
            # Save the sta_index for each array, as we will need it
            # later to match measurements to stations
            sta_indices[arrname] = data.field('sta_index')

    # Then get any science measurements
    for hdu in hdulist:
        header = hdu.header
        data = hdu.data
        if hdu.name in ('OI_VIS', 'OI_VIS2', 'OI_T3', 'AMBER_SPECTRUM'):
            if 'ARRNAME' in header.keys():
                arrname = header['ARRNAME']
            else:
                arrname = None
            if arrname and newobj.array:
                array = newobj.array[arrname]
            else:
                array = None
            wavelength = newobj.wavelength[header['INSNAME']]
        if hdu.name == 'OI_VIS':
            for row in data:
                date = getDate(hdu, hdulist)
                timeobs = _datetime.datetime(int(date[0]), int(date[1]), int(
                    date[2])) + _datetime.timedelta(seconds=_np.around(
                        row.field('TIME'), 2))
                int_time = row.field('INT_TIME')
                visamp = _np.reshape(row.field('VISAMP'), -1)
                visamperr = _np.reshape(row.field('VISAMPERR'), -1)
                visphi = _np.reshape(row.field('VISPHI'), -1)
                visphierr = _np.reshape(row.field('VISPHIERR'), -1)
                if 'CFLUX' in row.array.names:
                    cflux = _np.reshape(row.field('CFLUX'), -1)
                else:
                    cflux = None
                if 'CFLUXERR' in row.array.names:
                    cfluxerr = _np.reshape(row.field('CFLUXERR'), -1)
                else:
                    cfluxerr = None
                flag = _np.reshape(row.field('FLAG'), -1)
                ucoord = row.field('UCOORD')
                vcoord = row.field('VCOORD')
                target = targetmap[row.field('TARGET_ID')]
                if array:
                    sta_index = row.field('STA_INDEX')
                    s1 = array.station[sta_indices[arrname] == sta_index[0]][0]
                    s2 = array.station[sta_indices[arrname] == sta_index[1]][0]
                    station = [s1, s2]
                else:
                    station = [None, None]
                newobj.vis = _np.append(newobj.vis, OI_VIS(timeobs=timeobs, 
                    int_time=int_time, visamp=visamp, visamperr=visamperr, 
                    visphi=visphi, visphierr=visphierr, flag=flag, 
                    ucoord=ucoord, vcoord=vcoord, wavelength=wavelength, 
                    target=target, array=array, station=station, cflux=cflux,
                    cfluxerr=cfluxerr))
        elif hdu.name == 'OI_VIS2':
            for row in data:
                date = getDate(hdu, hdulist)
                timeobs = _datetime.datetime(int(date[0]), int(date[1]), int(
                    date[2])) + _datetime.timedelta(seconds=_np.around(
                        row.field('TIME'), 2))
                int_time = row.field('INT_TIME')
                vis2data = _np.reshape(row.field('VIS2DATA'), -1)
                vis2err = _np.reshape(row.field('VIS2ERR'), -1)
                flag = _np.reshape(row.field('FLAG'), -1)
                ucoord = row.field('UCOORD')
                vcoord = row.field('VCOORD')
                target = targetmap[row.field('TARGET_ID')]
                if array:
                    sta_index = row.field('STA_INDEX')
                    s1 = array.station[sta_indices[arrname] == sta_index[0]][0]
                    s2 = array.station[sta_indices[arrname] == sta_index[1]][0]
                    station = [s1, s2]
                else:
                    station = [None, None]
                newobj.vis2 = _np.append(newobj.vis2, OI_VIS2(timeobs=timeobs, 
                    int_time=int_time, vis2data=vis2data, vis2err=vis2err, 
                    flag=flag, ucoord=ucoord, vcoord=vcoord, 
                    wavelength=wavelength, target=target, array=array,
                    station=station))
        elif hdu.name == 'OI_T3':
            for row in data:
                date = getDate(hdu, hdulist)
                timeobs = _datetime.datetime(int(date[0]), int(date[1]), int(
                    date[2])) + _datetime.timedelta(seconds=_np.around(
                        row.field('TIME'), 2))
                int_time = row.field('INT_TIME')
                t3amp = _np.reshape(row.field('T3AMP'), -1)
                t3amperr = _np.reshape(row.field('T3AMPERR'), -1)
                t3phi = _np.reshape(row.field('T3PHI'), -1)
                t3phierr = _np.reshape(row.field('T3PHIERR'), -1)
                flag = _np.reshape(row.field('FLAG'), -1)
                u1coord = row.field('U1COORD')
                v1coord = row.field('V1COORD')
                u2coord = row.field('U2COORD')
                v2coord = row.field('V2COORD')
                target = targetmap[row.field('TARGET_ID')]
                if array:
                    sta_index = row.field('STA_INDEX')
                    s1 = array.station[sta_indices[arrname] == sta_index[0]][0]
                    s2 = array.station[sta_indices[arrname] == sta_index[1]][0]
                    s3 = array.station[sta_indices[arrname] == sta_index[2]][0]
                    station = [s1, s2, s3]
                else:
                    station = [None, None, None]
                newobj.t3 = _np.append(newobj.t3, OI_T3(timeobs=timeobs, 
                    int_time=int_time, t3amp=t3amp, t3amperr=t3amperr, 
                    t3phi=t3phi, t3phierr=t3phierr, flag=flag, u1coord=u1coord,
                    v1coord=v1coord, u2coord=u2coord, v2coord=v2coord, 
                    wavelength=wavelength, target=target, array=array, 
                    station=station))
        elif hdu.name == 'AMBER_SPECTRUM':
            if not quiet:
                print('AMBER SPEC INFO read')
            dataAS0 = []
            dataASf = []
            dataAS0err = []
            dataASferr = []
            for row in data:
                dataAS0 = _np.append(dataAS0, row['SPECTRUM'])
                dataAS0err = _np.append(dataAS0err, row['SPECTRUM'])
            dataAS0 = dataAS0.reshape(-1)
            dataAS0err = dataAS0err.reshape(-1)
            for i in range(3):
                dataASf = _np.append(dataASf, dataAS0[i::3])
                dataASferr = _np.append(dataASferr, dataAS0err[i::3])
            dataASf = dataASf.reshape(3, -1)
            dataASferr = dataASferr.reshape(3, -1)
            # datanames = data.names
            for i in range(3):
                # if 'EFF_BAND' in datanames:
                newobj.amberspec = _np.append(newobj.amberspec, 
                    AMBER_SPECTRUM(eff_wave=data.field('EFF_WAVE'), 
                        eff_band=data.field('EFF_BAND'), wavelength=wavelength, 
                        interfspec=data.field('INTERF_SPECTRUM'),
                        interfspecerr=data.field('INTERF_SPECTRUM_ERROR'),
                        spectrum=dataASf[i], spectrumerr=dataASferr[i]))

    hdulist.close()
    if not quiet:
        newobj.info(recursive=False)

    return newobj

# MAIN ###
if __name__ == "__main__":
    pass
